from enum import Enum


# Flow modifier parameters
class FunctionType(str, Enum):
    Constant: str = "constant"
    Linear: str = "linear"
    Exponential: str = "exponential"
    Sigmoid: str = "sigmoid"


class ChangeType(str, Enum):
    Value: str = "Value"
    Proportional: str = "%"


class FlowType(str, Enum):
    Absolute: str = "ABS"
    Relative: str = "REL"




import warnings
from typing import List, Union, Any, Dict
import numpy as np
import pandas as pd
from .datastructures import Process, Flow, Stock, FlowModifier, ScenarioDefinition, Color
from .parameters import ParameterName, ParameterFillMethod, StockDistributionType, ParameterScenarioType

# Suppress openpyxl warnings about Data Validation being suppressed
warnings.filterwarnings('ignore', category=UserWarning, module="openpyxl")
warnings.filterwarnings('ignore', category=FutureWarning, module="openpyxl")


class DataProvider(object):
    def __init__(self, filename: str = "",
                 sheet_settings_name: str = "Settings",
                 sheet_settings_col_range: Union[str, int] = "B:C",
                 sheet_settings_skip_num_rows: int = 5,
                 ):

        self._workbook = None
        self._param_name_to_value = {}
        self._processes: List[Process] = []
        self._flows: List[Flow] = []
        self._stocks: List[Stock] = []
        self._scenario_definitions: List[ScenarioDefinition] = []
        self._colors: List[Color] = []
        self._sheet_name_processes: Union[None, str] = None
        self._sheet_name_flows: Union[None, str] = None
        self._sheet_name_scenarios: Union[None, str] = None
        self._sheet_name_colors: Union[None, str] = None
        self._sheet_name_process_positions: Union[None, str] = None
        self._year_to_process_id_to_position = {}

        # Check that all required keys exists
        required_params = [
            [ParameterName.SheetNameProcesses,
             str,
             "Sheet name that contains data for Processes, (e.g. Processes)",
             ],
            [ParameterName.SkipNumRowsProcesses,
             int,
             "Number of rows to skip when reading data for Processes (e.g. 2). NOTE: Header row must be the first row to read!",
             ],
            [ParameterName.IgnoreColumnsProcesses,
             list,
             "Columns to ignore when reading Process sheet"
             ],

            # Flow related
            [ParameterName.SheetNameFlows,
             str,
             "Sheet name that contains data for Flows (e.g. Flows)",
             ],
            [ParameterName.SkipNumRowsFlows,
             int,
             "Number of rows to skip when reading data for Processes (e.g. 2). NOTE: Header row must be the first row to read!"
             ],
            [ParameterName.IgnoreColumnsFlows,
             list,
             "Columns to ignore when reading Flows sheet"
             ],

            # Model related
            [ParameterName.StartYear,
             int,
             "Starting year of the model"
             ],
            [ParameterName.EndYear,
             int,
             "Ending year of the model, included in time range"
             ],
            [ParameterName.DetectYearRange,
             bool,
             "Detect the year range automatically from file"
             ],
            [ParameterName.UseVirtualFlows,
             bool,
             "Use virtual flows (create missing flows for Processes that have imbalance of input and output flows, i.e. unreported flows)"
             ],
            [ParameterName.VirtualFlowsEpsilon,
             float,
             "Maximum allowed absolute difference of process input and outputs before creating virtual flow"
             ],
            [ParameterName.BaselineValueName,
             str,
             "Baseline value name. Name of the value type that is used as baseline e.g. 'Solid wood equivalent'"
             ],
            [ParameterName.BaselineUnitName,
             str,
             "Baseline unit name. This is used with relative flows when exporting flow data to CSVs."
             ],
        ]

        # Optional parameters entry structure:
        # Name of the parameter, expected value type, comment, default value
        optional_params = [
            [ParameterName.ConversionFactorCToCO2,
             float,
             "Conversion factor from C to CO2",
             None,
             ],

            [ParameterName.FillMissingAbsoluteFlows,
             bool,
             "Fill missing absolute flows with previous valid flow data?",
             True,
             ],

            [ParameterName.FillMissingRelativeFlows,
             bool,
             "Fill missing relative flows with previous valid flow data?",
             True,
             ],

            [ParameterName.FillMethod,
             str,
             "Fill method if either fill_missing_absolute_flows or fill_missing_relative_flows is enabled",
             ParameterFillMethod.Zeros,
             ],
            [ParameterName.UseScenarios,
             bool,
             "Run scenarios",
             True,
             ],
            [ParameterName.SheetNameScenarios,
             str,
             "Sheet name that contains data for scenarios (flow modifiers and constraints)",
             "Scenarios",
             ],
            [ParameterName.IgnoreColumnsScenarios,
             list,
             "Columns to ignore when reading Scenarios sheet",
             [],
             ],
            [ParameterName.ScenarioType,
             str,
             "Scenario type (Constrained / Unconstrained)",
             ParameterScenarioType.Constrained,
             # ParameterScenarioType.Unconstrained,
             ],
            [ParameterName.SheetNameColors,
             str,
             "Sheet name that contains data for transformation stage colors (e.g. Colors)",
             None,
             ],
            [ParameterName.IgnoreColumnsColors,
             list,
             "Columns to ignore when reading Colors sheet",
             [],
             ],
            [ParameterName.CreateNetworkGraphs,
             bool,
             "Create network graphs to visualize process connections for each scenario",
             False,
             ],
            [ParameterName.CreateSankeyCharts,
             bool,
             "Create Sankey charts for each scenario",
             True,
             ],
            [ParameterName.OutputPath,
             str,
             "Path to directory where all output is created (relative to running script)",
             "output",
             ],
            [ParameterName.ShowPlots,
             bool,
             "Show Matplotlib plots",
             True,
             ],
            [ParameterName.VisualizeInflowsToProcesses,
             list,
             "Create inflow visualization and export data for process IDs defined in here. " +
             "Each process ID must be separated by comma (',')",
             [],
             ],
            [ParameterName.PrioritizeLocations,
             list,
             "Prioritize process locations. If stock is outflowing to prioritized location, then "
             "ignore that amount as inflows to stock. This is to simulate trade flows happening during the timestep "
             "which should not go in to the stock",
             [],
             ],
            [ParameterName.PrioritizeTransformationStages,
             list,
             "Prioritize process transformation stages. If stock is outflowing to prioritized transformation stage, "
             "then ignore that amount as inflows to stock. This is to simulate trade flows happening during "
             "the timestep which should not go in to the stock",
             [],
             ],
            [ParameterName.SheetNameProcessPositions,
             str,
             "Sheet name that contains data for process positions in normalized format (position data in range [0,1])",
             None,
             ],
            [ParameterName.CheckErrors,
             bool,
             "Check errors when building data (development)",
             True,
             ],
            [ParameterName.IncludeMetadata,
             bool,
             "Include scenario metadata",
             # True,
             False,
             ],
        ]

        param_type_to_str = {int: "integer", float: "float", str: "string", bool: "boolean", list: "list"}

        # Read settings sheet from the file
        param_name_to_value = {}
        try:
            with pd.ExcelFile(filename) as xls:
                try:
                    sheet_settings = pd.read_excel(io=xls,
                                                   sheet_name=sheet_settings_name,
                                                   usecols=sheet_settings_col_range,
                                                   skiprows=sheet_settings_skip_num_rows,
                                                   )

                    for row_index, row in sheet_settings.iterrows():
                        param_name, param_value = row
                        param_name_to_value[param_name] = param_value

                except ValueError as e:
                    raise Exception("DataProvider: Settings sheet '{}' not found in file {}!".format(
                        sheet_settings_name, filename))

        except FileNotFoundError as ex:
            raise Exception("File not found: {}".format(filename))

        # Check that all required params are defined in settings sheet
        missing_params = []
        for entry in required_params:
            param_name, param_type, param_desc = entry
            if param_name not in param_name_to_value:
                missing_params.append(entry)

        # Print missing parameters and information
        if missing_params:
            print("DataProvider: Settings sheet (Sheet) is missing required following parameters:")
            max_param_name_len = 0
            for entry in missing_params:
                param_name = entry[0]
                max_param_name_len = len(param_name) if len(param_name) > max_param_name_len else max_param_name_len

            for entry in missing_params:
                param_name, param_type, param_desc = entry
                fixed_param_name = "{:" + str(max_param_name_len) + "}"
                fixed_param_name = fixed_param_name.format(param_name)
                print("\t{} (type: {}). {}".format(fixed_param_name, param_type_to_str[param_type], param_desc))

            raise Exception(-1)

        # Check that required and optionals parameters are correct types
        for entry in required_params:
            param_name, param_type, param_desc = entry
            if param_name in param_name_to_value:
                found_param_value = param_name_to_value[param_name]
                found_param_type = type(found_param_value)
                try:
                    if param_type is bool:
                        found_param_value = self._to_bool(found_param_value)

                    if param_type is list:
                        found_param_value = self._to_list(found_param_value)

                    found_param_value = param_type(found_param_value)
                    self._param_name_to_value[param_name] = found_param_value
                except ValueError as e:
                    print("Invalid type for required parameter '{}': expected {}, got {}".format(
                        param_name, param_type_to_str[param_type], param_type_to_str[found_param_type]))

        # Check that optional parameters are correct types
        for entry in optional_params:
            param_name, param_type, param_desc, param_default_value = entry
            if param_name in param_name_to_value:
                found_param_value = param_name_to_value[param_name]
                found_param_type = type(found_param_value)
                try:
                    if param_type is bool:
                        found_param_value = self._to_bool(found_param_value)

                    if param_type is list:
                        found_param_value = self._to_list(found_param_value)

                    self._param_name_to_value[param_name] = param_type(found_param_value)
                except ValueError as e:
                    print("Invalid type for optional parameter '{}': expected {}, got {}".format(
                        param_name, param_type_to_str[param_type], param_type_to_str[found_param_type]))

                # Check that FillMethod has valid value
                if param_name is ParameterName.FillMethod:
                    # Convert found param name and valid fill method names to lowercase
                    # and check if found param name is one of the valid method names
                    valid_fill_method_names = [fill_method_name for fill_method_name in ParameterFillMethod]
                    found_param_value_lower = found_param_value.lower().strip()
                    valid_method_names_lower = [name.lower().strip() for name in valid_fill_method_names]
                    if found_param_value_lower in valid_method_names_lower:
                        # Get the actual parameter name from ParameterFillMethod-enum
                        fill_method_index = valid_method_names_lower.index(found_param_value_lower)
                        found_param_value = valid_fill_method_names[fill_method_index]
                        self._param_name_to_value[param_name] = found_param_value
                    else:
                        print("{} not valid value for {}! ".format(found_param_value, param_name), end="")
                        print("Valid values are: ", end="")
                        for index, method_name in enumerate(valid_fill_method_names):
                            print(method_name, end="")
                            if index < len(valid_fill_method_names) - 1:
                                print(", ", end="")
                            else:
                                print("")

                        self._param_name_to_value[param_name] = param_default_value
                        print("")
                        raise Exception(-1)

                elif param_name is ParameterName.ScenarioType:
                    valid_scenario_type_names = [scenario_type_name for scenario_type_name in ParameterScenarioType]
                    found_param_value_lower = found_param_value.lower().strip()
                    valid_method_names_lower = [name.lower().strip() for name in valid_scenario_type_names]
                    if found_param_value_lower in valid_method_names_lower:
                        # Get the actual parameter name from ParameterFillMethod-enum
                        scenario_type_index = valid_method_names_lower.index(found_param_value_lower)
                        found_param_value = valid_scenario_type_names[scenario_type_index]
                        self._param_name_to_value[param_name] = found_param_value
                    else:
                        print("{} not valid value for {}! ".format(found_param_value, param_name), end="")
                        print("Valid values are: ", end="")
                        for index, method_name in enumerate(valid_scenario_type_names):
                            print(method_name, end="")
                            if index < len(valid_scenario_type_names) - 1:
                                print(", ", end="")
                            else:
                                print("")

                        self._param_name_to_value[param_name] = param_default_value
                        print("")
                        raise Exception(-1)

            else:
                # Use default optional parameter value
                self._param_name_to_value[param_name] = param_default_value

        # ********************************************
        # * Read processes and flows from Excel file *
        # ********************************************

        # Create Processes and Flows
        sheet_name_processes = self._param_name_to_value.get(ParameterName.SheetNameProcesses, None)
        ignore_columns_processes = self._param_name_to_value.get(ParameterName.IgnoreColumnsProcesses, [])
        skip_num_rows_processes = self._param_name_to_value.get(ParameterName.SkipNumRowsProcesses, None)

        sheet_name_flows = self._param_name_to_value.get(ParameterName.SheetNameFlows, None)
        ignore_columns_flows = self._param_name_to_value.get(ParameterName.IgnoreColumnsFlows, [])
        skip_num_rows_flows = self._param_name_to_value.get(ParameterName.SkipNumRowsFlows, None)

        sheet_name_scenarios = self._param_name_to_value.get(ParameterName.SheetNameScenarios, None)
        ignore_columns_scenarios = self._param_name_to_value.get(ParameterName.IgnoreColumnsScenarios, [])
        skip_num_rows_scenarios = self._param_name_to_value.get(ParameterName.SkipNumRowsScenarios, None)

        # Optional
        # Colors
        sheet_name_colors = self._param_name_to_value.get(ParameterName.SheetNameColors, None)
        ignore_columns_colors = self._param_name_to_value.get(ParameterName.IgnoreColumnsColors, [])
        skip_num_rows_colors = self._param_name_to_value.get(ParameterName.SkipNumRowsColors, None)

        # Process positions
        sheet_name_process_positions = self._param_name_to_value.get(ParameterName.SheetNameProcessPositions, None)

        use_scenarios = self._param_name_to_value[ParameterName.UseScenarios]
        if not use_scenarios:
            sheet_name_scenarios = ""

        # Include metadata
        include_metadata = self._param_name_to_value[ParameterName.IncludeMetadata]

        # Sheet name to DataFrame
        sheets = {}
        try:
            with pd.ExcelFile(filename) as xls:
                try:
                    sheet_processes = pd.read_excel(xls,
                                                    sheet_name=sheet_name_processes,
                                                    skiprows=skip_num_rows_processes)
                    sheets[sheet_name_processes] = self._drop_ignored_columns_from_sheet(sheet_processes,
                                                                                         ignore_columns_processes)
                except ValueError:
                    pass

                try:
                    sheet_flows = pd.read_excel(xls,
                                                sheet_name=sheet_name_flows,
                                                skiprows=skip_num_rows_flows)
                    sheets[sheet_name_flows] = self._drop_ignored_columns_from_sheet(sheet_flows,
                                                                                     ignore_columns_flows)
                except ValueError:
                    pass

                # Optionals
                if use_scenarios:
                    try:
                        sheet_scenarios = pd.read_excel(xls,
                                                        sheet_name=sheet_name_scenarios,
                                                        skiprows=skip_num_rows_scenarios)
                        sheets[sheet_name_scenarios] = self._drop_ignored_columns_from_sheet(sheet_scenarios,
                                                                                             ignore_columns_scenarios)

                    except ValueError:
                        pass

                try:
                    sheet_colors = pd.read_excel(xls,
                                                 sheet_name=sheet_name_colors,
                                                 skiprows=skip_num_rows_colors)
                    sheets[sheet_name_colors] = self._drop_ignored_columns_from_sheet(sheet_colors,
                                                                                      ignore_columns_colors)
                except ValueError:
                    pass

                try:
                    sheet_process_positions = pd.read_excel(xls, sheet_name=sheet_name_process_positions)
                    sheets[sheet_name_process_positions] = sheet_process_positions
                except ValueError:
                    pass

        except FileNotFoundError:
            raise Exception("Settings file '{}' not found".format(filename))

        # Check that all the required sheets exists
        required_sheet_names = [sheet_name_processes, sheet_name_flows]
        missing_sheet_names = self._check_missing_sheet_names(required_sheet_names, sheets)
        if missing_sheet_names:
            print("Settings file '{}' is missing following required sheets:".format(filename))
            for key in missing_sheet_names:
                print("\t- {}".format(key))
            raise Exception(-1)

        self._sheet_name_processes = sheet_name_processes
        self._sheet_name_flows = sheet_name_flows

        # Check that all optional sheets exists (only if optional sheets have been defined)
        optional_sheet_names = [sheet_name_colors, sheet_name_process_positions]

        if sheet_name_scenarios:
            optional_sheet_names.append(sheet_name_scenarios)

        missing_sheet_names = self._check_missing_sheet_names(optional_sheet_names, sheets)
        if missing_sheet_names:
            errors = []
            s = "Settings file '{}' is missing following sheets:".format(filename)
            errors.append(s)
            for key in missing_sheet_names:
                s = "\t- {}".format(key)
                errors.append(s)
            raise Exception(errors)

        self._sheet_name_scenarios = sheet_name_scenarios
        self._sheet_name_colors = sheet_name_colors
        self._sheet_name_process_positions = sheet_name_process_positions

        # Create Processes
        rows_processes = []
        df_processes = sheets[self._sheet_name_processes]
        for (row_index, row) in df_processes.iterrows():
            row = self._convert_row_nan_to_none(row)
            rows_processes.append(row)

        self._processes = self._create_objects_from_rows(Process,
                                                         rows_processes,
                                                         row_start=skip_num_rows_processes)

        # Create Flows
        rows_flows = []
        df_flows = sheets[self._sheet_name_flows]
        for (row_index, row) in df_flows.iterrows():
            row = self._convert_row_nan_to_none(row)
            rows_flows.append(row)

        self._flows = self._create_objects_from_rows(Flow,
                                                     rows_flows,
                                                     row_start=skip_num_rows_flows)

        # Create Stocks from Processes
        self._stocks = self._create_stocks_from_processes(self._processes)

        # Create alternative scenarios (optional)
        if sheet_name_scenarios:
            rows_scenarios = []
            df_scenarios = sheets[self._sheet_name_scenarios]
            for (row_index, row) in df_scenarios.iterrows():
                row = self._convert_row_nan_to_none(row)
                rows_scenarios.append(row)

            self._scenario_definitions = self._create_scenario_definitions(rows_scenarios)

        # Create colors (optional)
        if sheet_name_colors:
            rows_colors = []
            df_colors = sheets[self._sheet_name_colors]
            for (row_index, row) in df_colors.iterrows():
                row = self._convert_row_nan_to_none(row)
                rows_colors.append(row)

            self._colors = self._create_colors(rows_colors)

        # Read and update detailed Process positions (optional)
        if sheet_name_process_positions:

            # Map year -> Process ID -> position
            year_to_process_id_to_position = {}
            df_process_positions = sheets[sheet_name_process_positions]
            for (row_index, row) in df_process_positions.iterrows():
                year = row.iloc[0]
                process_id = row.iloc[1]
                x = self._to_normalized_float(row.iloc[2], 3)
                y = self._to_normalized_float(row.iloc[3], 3)
                normalized_x = np.clip(x, 0.001, 0.999)
                normalized_y = np.clip(y, 0.001, 0.999)

                if year not in year_to_process_id_to_position:
                    year_to_process_id_to_position[year] = {}

                process_id_to_position = year_to_process_id_to_position[year]
                if process_id not in process_id_to_position:
                    process_id_to_position[process_id] = [normalized_x, normalized_y]

            self._year_to_process_id_to_position = year_to_process_id_to_position

    @property
    def sheet_name_processes(self):
        return self._sheet_name_processes

    @property
    def sheet_name_flows(self):
        return self._sheet_name_flows

    def _check_missing_sheet_names(self, required_sheet_names: List[str], sheets: Dict[str, pd.DataFrame]):
        missing_sheet_names = []
        for key in required_sheet_names:
            if key not in sheets:
                missing_sheet_names.append(key)

        return missing_sheet_names

    def _create_objects_from_rows(self, object_type=None, rows=None, row_start=-1) -> List:
        if rows is None:
            rows = []

        result = []
        if not object_type:
            return result

        row_number = row_start + 2
        for row in rows:
            if not self._is_row_valid(row):
                row_number += 1
                continue

            new_instance = object_type(row, row_number)
            if new_instance.is_valid():
                result.append(new_instance)

            row_number += 1

        return result

    def _create_stocks_from_processes(self, processes=None) -> List[Stock]:
        # Create stocks only for Processes that have lifetime > 1
        # NOTE: If stock type is LandfillDecay* then set lifetime to 1 by default
        # if stock lifetime is not defined
        if processes is None:
            processes = []

        # Ignore stock lifetime if stock distribution type is any of these
        ignore_stock_lifetime_for_types = {StockDistributionType.LandfillDecayWood,
                                           StockDistributionType.LandfillDecayPaper}
        result = []
        for process in processes:
            ignore_stock_lifetime = process.stock_distribution_type in ignore_stock_lifetime_for_types

            # Add stock lifetime if stock type is in ignore_stock_lifetime
            # This is used for LandfillDecay* stock types and stock lifetime is needed
            # so that ODYM is able to calculate decays properly
            if ignore_stock_lifetime and process.stock_lifetime == 0:
                process.stock_lifetime = 1

            if process.stock_lifetime == 0:
                continue

            new_stock = Stock(process, row_number=process.row_number)
            if new_stock.is_valid():
                result.append(new_stock)

        return result

    def _create_scenario_definitions(self, rows: List[Any] = None) -> List[ScenarioDefinition]:
        if not rows:
            rows = []

        flow_modifiers = []
        for row_index, row in enumerate(rows):
            new_flow_modifier = FlowModifier(row)
            new_flow_modifier.row_number = row_index + 2  # Header = row 1
            if new_flow_modifier.is_valid():
                flow_modifiers.append(new_flow_modifier)

        # Build scenario mappings
        result = []
        if not flow_modifiers:
            # No alternative scenarios found, create later only the baseline scenario
            pass
        else:
            # Map scenario names to flow modifiers
            scenario_name_to_flow_modifiers = {}
            for flow_modifier in flow_modifiers:
                scenario_name = flow_modifier.scenario_name
                if scenario_name not in scenario_name_to_flow_modifiers:
                    scenario_name_to_flow_modifiers[scenario_name] = []
                scenario_name_to_flow_modifiers[scenario_name].append(flow_modifier)

            # Create scenario definitions from mappings
            for scenario_name, scenario_flow_modifiers in scenario_name_to_flow_modifiers.items():
                new_scenario_definition = ScenarioDefinition(scenario_name, scenario_flow_modifiers)
                result.append(new_scenario_definition)
        return result

    def _create_colors(self, rows: List[Any] = None) -> List[Color]:
        if not rows:
            rows = []

        result = []
        for row_index, row in enumerate(rows):
            # Header = row 1
            result.append(Color(row, row_number=row_index + 2))
        return result

    def _is_row_valid(self, row):
        """
        Check if row is valid.
        Defining all first 4 columns makes the row valid.

        :param row: Target row
        :return: True if row is valid, false otherwise
        """
        # Each row must have all first columns defined
        cols = row.iloc[0:4]
        if any(pd.isna(cols)):
            return False

        return True

    def _convert_row_nan_to_none(self, row: pd.Series) -> pd.Series:
        """
        Check the row and convert NaN to None.
        Modifies the original row.

        :param row: Series
        :return: Returns the original modified row
        """
        for col_name, value in row.items():
            if np.isreal(value) and np.isnan(value):
                row[col_name] = None
        return row

    def _to_bool(self, value: Any) -> bool:
        """
        Check and convert value to bool.
        If value is string then converts to lowercase and checks if value is either "true" or "false"
        and returns corresponding value as bool.\n
        NOTE: Only converts value to string and checks bool validity, no other checking is done.

        :param value: Value to convert to bool
        :return: Value as bool
        """
        if isinstance(value, str):
            if value.lower() == "true":
                return True

        if isinstance(value, bool):
            return value

        else:
            return False

    def _to_normalized_float(self, val: Any, decimals: int = 3) -> float:
        """
        Check and convert val to normalized float [0, 1] with optional precision.
        Returns -1 if conversion fails.

        :param val: Target value
        :param decimals: Number of decimals to keep (default = 3)
        :return: Float
        """
        result = -1.0
        try:
            result = round(float(val), decimals)
        except ValueError as ex:
            pass

        return result

    def _to_list(self, value: Any, sep=',', allowed_chars = [":"]) -> List[str]:
        """
        Check and convert value to list of strings.
        Default separator is comma (',')
        Returns empty list of conversion is not possible.

        :param value: Value to be converted to list
        :return: List of strings
        """

        result = []
        if type(value) is not str:
            # float and int are not valid types, just return empty list
            result = []
        else:
            # Split the string by sep and ignore all elements that are not strings
            splits = value.split(sep)
            for s in splits:
                stripped = str(s).strip()
                has_allowed_char = np.any([stripped.find(c) < 0 for c in allowed_chars])
                if not stripped.isalpha() and has_allowed_char:
                    continue

                result.append(stripped)

        return result

    def _excel_column_name_to_index(self, col_name: str) -> int:
        n = 0
        for c in col_name:
            n = \
                (n * 26) + 1 + ord(c) - ord('A')
        return n

    def _excel_column_index_to_name(self, col_index: int) -> str:
        name = ''
        n = col_index
        while n > 0:
            n, r = divmod(n - 1, 26)
            name = chr(r + ord('A')) + name
        return name

    def _drop_ignored_columns_from_sheet(self, sheet: pd.DataFrame, ignored_col_names: List[str]) -> pd.DataFrame:
        """
        Drop list of Excel column names from DataFrame.
        Does not modify original pd.DataFrame.

        :param sheet: Target pd.DataFrame
        :param ignored_col_names: List of ignored Excel column names, e.g. ['A', 'B', 'E']
        :return:
        """

        # Drop ignored columns
        col_indices_to_drop = []
        for ignored_col_name in ignored_col_names:
            # Convert index of Excel column name (1-based) to 0-based index for DataFrame
            col_index = self._excel_column_name_to_index(ignored_col_name) - 1
            col_indices_to_drop.append(col_index)

        new_sheet = sheet.drop(sheet.columns[col_indices_to_drop], axis=1)
        return new_sheet

    def get_model_params(self) -> dict[ParameterName, Any]:
        """
        Get model parameters read from the data file.

        :return: Dictionary of parameter name to parameter value
        """
        return self._param_name_to_value

    def get_processes(self) -> List[Process]:
        """
        Get all Processes.

        :return: List of Processes
        """
        return self._processes

    def get_flows(self) -> List[Flow]:
        """
        Get all Flows.

        :return: List of Flows
        """
        return self._flows

    def get_stocks(self) -> List[Stock]:
        """
        Get all Stocks.

        :return: List of Stocks
        """
        return self._stocks

    def get_scenario_definitions(self) -> List[ScenarioDefinition]:
        """
        Get all ScenarioDefinitions.

        :return: List of ScenarioDefinitions
        """
        return self._scenario_definitions

    def get_color_definitions(self) -> List[Color]:
        return self._colors

    def get_process_positions(self) -> Dict[int, Dict[str, List[float]]]:
        """
        Get year -> Process ID -> position mappings.

        :return: Dictionary (year -> Dictionary(Process ID, List[x, y]))
        """
        return self._year_to_process_id_to_position



# -*- coding: utf-8 -*-
"""
Created on Mon Aug 11 16:19:39 2014

"""
import os
import sys
import imp
# Put location of 
sys.path.insert(0,  os.path.abspath(os.path.join(os.path.dirname(__file__), '..\\..')) + '\\modules') # add ODYM module directory to system path
#NOTE: Hidden variable __file__ must be know to script for the directory structure to work.
# Therefore: When first using the model, run the entire script with F5 so that the __file__ variable can be created.

import dynamic_stock_model as dsm # remove and import the class manually if this unit test is run as standalone script
imp.reload(dsm)

import numpy as np
import unittest
import scipy


###############################################################################
"""My Input for fixed lifetime"""
Time_T_FixedLT = np.arange(0,10)
Time_T_30      = np.arange(0,30)
Inflow_T_FixedLT  = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
lifetime_FixedLT  = {'Type': 'Fixed', 'Mean': np.array([5])}
lifetime_FixedLT0 = {'Type': 'Fixed', 'Mean': np.array([0])}
#lifetime_FixedLT = {'Type': 'Fixed', 'Mean': np.array([5,5,5,5,5,5,5,5,5,5])}
lifetime_NormLT  = {'Type': 'Normal', 'Mean': np.array([5]), 'StdDev': np.array([1.5])}
lifetime_NormLT0 = {'Type': 'Normal', 'Mean': np.array([0]), 'StdDev': np.array([1.5])}
lifetime_NormLT8 = {'Type': 'Normal', 'Mean': np.array([8]), 'StdDev': np.array([3])}
###############################################################################
"""My Output for fixed lifetime"""
Outflow_T_FixedLT = np.array([0, 0, 0, 0, 0, 1, 2, 3, 4, 5])

Outflow_TC_FixedLT = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                               [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                               [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                               [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                               [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                               [1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                               [0, 2, 0, 0, 0, 0, 0, 0, 0, 0],
                               [0, 0, 3, 0, 0, 0, 0, 0, 0, 0],
                               [0, 0, 0, 4, 0, 0, 0, 0, 0, 0],
                               [0, 0, 0, 0, 5, 0, 0, 0, 0, 0]])

Stock_T_FixedLT = np.array([1, 3, 6, 10, 15, 20, 25, 30, 35, 40])

StockChange_T_FixedLT = np.array([1, 2, 3, 4, 5, 5, 5, 5, 5, 5])

Stock_TC_FixedLT = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                             [1, 2, 0, 0, 0, 0, 0, 0, 0, 0],
                             [1, 2, 3, 0, 0, 0, 0, 0, 0, 0],
                             [1, 2, 3, 4, 0, 0, 0, 0, 0, 0],
                             [1, 2, 3, 4, 5, 0, 0, 0, 0, 0],
                             [0, 2, 3, 4, 5, 6, 0, 0, 0, 0],
                             [0, 0, 3, 4, 5, 6, 7, 0, 0, 0],
                             [0, 0, 0, 4, 5, 6, 7, 8, 0, 0],
                             [0, 0, 0, 0, 5, 6, 7, 8, 9, 0],
                             [0, 0, 0, 0, 0, 6, 7, 8, 9, 10]])

Bal    = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
Bal9   = np.zeros((9))
Bal30  = np.zeros((30))
"""My Output for normally distributed lifetime"""

Stock_TC_NormLT = np.array([[  9.99570940e-01,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00],
       [  9.96169619e-01,   1.99914188e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00],
       [  9.77249868e-01,   1.99233924e+00,   2.99871282e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00],
       [  9.08788780e-01,   1.95449974e+00,   2.98850886e+00,
          3.99828376e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00],
       [  7.47507462e-01,   1.81757756e+00,   2.93174960e+00,
          3.98467848e+00,   4.99785470e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00],
       [  5.00000000e-01,   1.49501492e+00,   2.72636634e+00,
          3.90899947e+00,   4.98084810e+00,   5.99742564e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00],
       [  2.52492538e-01,   1.00000000e+00,   2.24252239e+00,
          3.63515512e+00,   4.88624934e+00,   5.97701772e+00,
          6.99699658e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00],
       [  9.12112197e-02,   5.04985075e-01,   1.50000000e+00,
          2.99002985e+00,   4.54394390e+00,   5.86349921e+00,
          6.97318734e+00,   7.99656752e+00,   0.00000000e+00,
          0.00000000e+00],
       [  2.27501319e-02,   1.82422439e-01,   7.57477613e-01,
          2.00000000e+00,   3.73753731e+00,   5.45273268e+00,
          6.84074908e+00,   7.96935696e+00,   8.99613846e+00,
          0.00000000e+00],
       [  3.83038057e-03,   4.55002639e-02,   2.73633659e-01,
          1.00997015e+00,   2.50000000e+00,   4.48504477e+00,
          6.36152146e+00,   7.81799894e+00,   8.96552657e+00,
          9.99570940e+00]])

Stock_T_NormLT = np.array([  0.99957094,   2.9953115 ,   5.96830193,   9.85008113,
        14.4793678 ,  19.60865447,  24.99043368,  30.46342411,
        35.95916467,  41.45873561])

Stock_SDM_NegInflow_NormLT    = np.array([0.0,0,1,2,3,5,6,8,8,6,6,6,6,6,8,8,8,9,10,12,9,9,9,9,9,9,9,9.0,9,9])
Stock_SDM_NegInflow_NormLTNeg = np.array([0.0,0,0,0,0,0,0,0,8,6,6,6,6,6,3,3,3,3,4,5,6,7,8,9,9,9,7.1,8,8.0,8])
Stock_SDM_PosInflow_NormLTNeg = np.array([0.0,0,0,0,0,0,0,0,8,8,8,8.2,9,9,8.5,8,7.5,7,6.5,6,5.5,7,8,9,8.8,8.5,8.1,8,8.0,8])

Outflow_T_NormLT = np.array([  4.29060333e-04,   4.25944090e-03,   2.70095728e-02,
         1.18220793e-01,   3.70713330e-01,   8.70713330e-01,
         1.61822079e+00,   2.52700957e+00,   3.50425944e+00,
         4.50042906e+00])

Outflow_TC_NormLT = np.array([[  4.29060333e-04,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00],
       [  3.40132023e-03,   8.58120666e-04,  -0.00000000e+00,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00],
       [  1.89197514e-02,   6.80264047e-03,   1.28718100e-03,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00],
       [  6.84610878e-02,   3.78395028e-02,   1.02039607e-02,
          1.71624133e-03,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00],
       [  1.61281318e-01,   1.36922176e-01,   5.67592541e-02,
          1.36052809e-02,   2.14530167e-03,  -0.00000000e+00,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00],
       [  2.47507462e-01,   3.22562636e-01,   2.05383263e-01,
          7.56790055e-02,   1.70066012e-02,   2.57436200e-03,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00],
       [  2.47507462e-01,   4.95014925e-01,   4.83843953e-01,
          2.73844351e-01,   9.45987569e-02,   2.04079214e-02,
          3.00342233e-03,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00],
       [  1.61281318e-01,   4.95014925e-01,   7.42522387e-01,
          6.45125271e-01,   3.42305439e-01,   1.13518508e-01,
          2.38092416e-02,   3.43248267e-03,  -0.00000000e+00,
         -0.00000000e+00],
       [  6.84610878e-02,   3.22562636e-01,   7.42522387e-01,
          9.90029850e-01,   8.06406589e-01,   4.10766527e-01,
          1.32438260e-01,   2.72105619e-02,   3.86154300e-03,
         -0.00000000e+00],
       [  1.89197514e-02,   1.36922176e-01,   4.83843953e-01,
          9.90029850e-01,   1.23753731e+00,   9.67687907e-01,
          4.79227614e-01,   1.51358011e-01,   3.06118821e-02,
          4.29060333e-03]])

StockChange_T_NormLT = np.array([ 0.99957094,  1.99574056,  2.97299043,  3.88177921,  4.62928667,
        5.12928667,  5.38177921,  5.47299043,  5.49574056,  5.49957094])

"""My Output for Weibull-distributed lifetime"""

Stock_TC_WeibullLT = np.array([[1, 0,  0,  0,  0,  0,  0, 0, 0,  0],  # computed with Excel and taken from there
                             [0.367879441,	2,	0,	0,	0,	0,	0,	0,	0,	0],
                             [0.100520187,	0.735758882,	3,	0,	0,	0,	0,	0,	0,	0],
                            [0.023820879,	0.201040373,	1.103638324,	4,	0,	0,	0,	0,	0,	0],
                            [0.005102464,	0.047641758,	0.30156056,	1.471517765,5,	0,	0,	0,	0,	0],
                            [0.001009149,	0.010204929,	0.071462637,	0.402080746,1.839397206,	6,	0,	0,	0,	0],
                            [0.000186736,	0.002018297,	0.015307393,	0.095283516,	0.502600933,	2.207276647,	7,	0,	0,	0],
                            [3.26256E-05,	0.000373472,	0.003027446,	0.020409858,	0.119104394,	0.60312112,	2.575156088,	8,	0,	0],
                            [5.41828E-06,	6.52513E-05,	0.000560208,	0.004036594,	0.025512322,	0.142925273,	0.703641306,	2.943035529,	9,	0],
                            [8.59762E-07,	1.08366E-05,	9.78769E-05,	0.000746944,	0.005045743,	0.030614786,	0.166746152,	0.804161493,	3.310914971,	10]])

Stock_T_WeibullLT = np.array([1,2.367879441,3.836279069,5.328499576,6.825822547,8.324154666,9.822673522,11.321225,12.8197819,14.31833966])

Outflow_T_WeibullLT = np.array([0,0.632120559,1.531600372,2.507779493,3.502677029,4.50166788,5.501481144,6.501448519,7.5014431,8.501442241])

Outflow_TC_WeibullLT = np.array([[0,	0,	0,	0,	0,	0,	0,	0,	0,	0],
                                [0.632120559,	0,	0,	0,	0,	0,	0,	0,	0,	0],
                                [0.267359255,	1.264241118,	0,	0,	0,	0,	0,	0,	0,	0],
                                [0.076699308,	0.534718509,	1.896361676,	0,	0,	0,	0,	0,	0,	0],
                                [0.018718414,	0.153398615,	0.802077764,	2.528482235,	0,	0,	0,	0,	0,	0],
                                [0.004093316,	0.037436829,	0.230097923,	1.069437018,	3.160602794,	0,	0,	0,	0,	0],
                                [0.000822413,	0.008186632,	0.056155243,	0.306797231,	1.336796273,	3.792723353,	0,	0,	0,	0],
                                [0.00015411,	0.001644825,	0.012279947,	0.074873658,	0.383496539,	1.604155527,	4.424843912,	0,	0,	0],
                                [2.72074E-05,	0.000308221,	0.002467238,	0.016373263,	0.093592072,	0.460195846,	1.871514782,	5.056964471,	0,	0],
                                [4.55852E-06, 5.44147E-05	,     0.000462331	,     0.00328965,   	0.020466579,	0.112310487,	0.536895154,	2.138874037,	5.689085029,	0]])

StockChange_T_WeibullLT = np.array([1,1.367879441,1.468399628,1.492220507,1.497322971,1.49833212,1.498518856,1.498551481,1.4985569,1.498557759])

lifetime_WeibullLT = {'Type': 'Weibull', 'Shape': np.array([1.2]), 'Scale': np.array([1])}
InitialStock_WB = np.array([0.01, 0.01, 0.08, 0.2,  0.2,  2,  2,  3,  4,  7.50])
Inflow_WB = np.array([11631.1250671964,	1845.6048709861,	2452.0593141014,	1071.0305279511,	198.1868742385,	391.9674590243,	83.9599583940,	29.8447516023,	10.8731273138,	7.5000000000])
# We need 10 digits AFTER the . to get a 9 digits after the . overlap with np.testing.
# The total number of counting digits is higher, because there are up to 5 digits before the .

# For the stock-driven model with initial stock, colculated with Excel
Sc_InitialStock_2_Ref = np.array([[ 3.29968072,  0.        ,  0.        ,  0.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  0.        ],
       [ 3.28845263,  5.1142035 ,  0.        ,  0.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  0.        ],
       [ 3.2259967 ,  5.09680099,  2.0068288 ,  0.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  0.        ],
       [ 3.        ,  5.        ,  2.        ,  4.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  0.        ],
       [ 2.46759471,  4.64972578,  1.962015  ,  3.98638888,  4.93427563,
         0.        ,  0.        ,  0.        ,  0.        ],
       [ 1.65054855,  3.82454624,  1.82456634,  3.91067739,  4.91748538,
         3.8721761 ,  0.        ,  0.        ,  0.        ],
       [ 0.83350238,  2.55819937,  1.50076342,  3.63671549,  4.82409004,
         3.85899993,  2.78772936,  0.        ,  0.        ],
       [ 0.30109709,  1.2918525 ,  1.00384511,  2.9913133 ,  4.48613916,
         3.78570788,  2.77824333,  3.36180162,  0.        ],
       [ 0.07510039,  0.46667297,  0.5069268 ,  2.00085849,  3.68999109,
         3.5205007 ,  2.72547754,  3.35036215,  3.66410986]])

Sc_InitialStock_2_Ref_Sum = np.array([  3.29968072,   8.40265614,  10.32962649,  14.        ,
        18.        ,  20.        ,  20.        ,  20.        ,  20.        ])

Oc_InitialStock_2_Ref = np.array([[  1.41636982e-03,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
          0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
       [  1.12280883e-02,   2.19524375e-03,  -0.00000000e+00,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00],
       [  6.24559363e-02,   1.74025106e-02,   8.61420234e-04,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00],
       [  2.25996698e-01,   9.68009922e-02,   6.82879736e-03,
          1.71697802e-03,  -0.00000000e+00,  -0.00000000e+00,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00],
       [  5.32405289e-01,   3.50274224e-01,   3.79849998e-02,
          1.36111209e-02,   2.11801070e-03,  -0.00000000e+00,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00],
       [  8.17046165e-01,   8.25179532e-01,   1.37448656e-01,
          7.57114903e-02,   1.67902556e-02,   1.66211031e-03,
         -0.00000000e+00,  -0.00000000e+00,  -0.00000000e+00],
       [  8.17046165e-01,   1.26634687e+00,   3.23802924e-01,
          2.73961897e-01,   9.33953405e-02,   1.31761643e-02,
          1.19661751e-03,  -0.00000000e+00,  -0.00000000e+00],
       [  5.32405289e-01,   1.26634687e+00,   4.96918311e-01,
          6.45402188e-01,   3.37950879e-01,   7.32920558e-02,
          9.48603036e-03,   1.44303487e-03,  -0.00000000e+00],
       [  2.25996698e-01,   8.25179532e-01,   4.96918311e-01,
          9.90454815e-01,   7.96148072e-01,   2.65207178e-01,
          5.27657861e-02,   1.14394721e-02,   1.57279902e-03]])

I_InitialStock_2_Ref  = np.array([ 3.30109709,  5.11639875,  2.00769022,  4.00171698,  4.93639364,  3.87383821,  2.78892598,  3.36324466,  3.66568266])

""" Test case with fixed lifetime for initial stock"""
Time_T_FixedLT_X = np.arange(1, 9, 1)
lifetime_FixedLT_X = {'Type': 'Fixed', 'Mean': np.array([5])}
InitialStock_X = np.array([0, 0, 0, 7, 5, 4, 3, 2])
Inflow_X = np.array([0, 0, 0, 7, 5, 4, 3, 2])

Time_T_FixedLT_XX = np.arange(1, 11, 1)
lifetime_NormLT_X = {'Type': 'Normal', 'Mean': np.array([5]), 'StdDev': np.array([1.5])}
InitialStock_XX = np.array([0.01, 0.01, 0.08, 0.2,  0.2,  2,  2,  3,  4,  7.50])
Inflow_XX = np.array([ 2.61070664,  0.43955789,  0.87708508,  0.79210262,  0.4,
        2.67555857,  2.20073139,  3.06983925,  4.01538044,  7.50321933])

""" Test case with normally distributed lifetime for initial stock and stock-driven model"""
Time_T_FixedLT_2  = np.arange(1, 10, 1)
Time_T_FixedLT_3  = np.arange(0, 30, 1)
lifetime_NormLT_2 = {'Type': 'Normal', 'Mean': np.array([5]), 'StdDev': np.array([1.5])}
lifetime_NormLT_3 = {'Type': 'LogNormal', 'Mean': np.array([6]), 'StdDev': np.array([3])}
lifetime_NormLT_4 = {'Type': 'Normal', 'Mean': np.array([7]), 'StdDev': np.array([2])}
InitialStock_2    = np.array([3,5,2,4])
InitialStock_8    = np.array([0.5,0.5,0.2,0.4,1,1,1,1])
FutureStock_2     = np.array([0.0,0.0,0.0,0.0,18.0,20,20,20,20.0])
ThisSwitchTime    = 5 # First year with future stock curve, start counting from 1.
Inflow_2          = np.array([3.541625588,	5.227890554,2.01531097,4])

""" Test case with lognormally distributed lifetime for inflow-driven model"""
# NOTE: for dynamic MFA the parameters average lifetime and standard deviation refer to the pdf curve as plotted, 
# whereas in the scipy and Excel functions, the parameters mu and sigma
# refer to the parameters of the underlying normal distribution.
# The parameters need to be converted as follows:
# for single (fixed) lifetime
Time_T_LN  = np.arange(0, 300, 1)
mu    = 20
sg    = 20
LT_LN = np.log(mu / np.sqrt(1 + mu * mu / (sg * sg))) # calculate parameter mu    of underlying normal distribution
SG_LN = np.sqrt(np.log(1 + mu * mu / (sg * sg)))      # calculate parameter sigma of underlying normal distribution
lifetime_LogNorm = {'Type': 'LogNormal', 'Mean': np.array([mu]), 'StdDev': np.array([sg])} # call with ACTUAL mu and sigma of curve,
# dsm makes conversion
SF = scipy.stats.lognorm.sf(x=Time_T_LN, s=SG_LN, loc = 0, scale=np.exp(LT_LN)) # values chosen according to description on
# https://docs.scipy.org/doc/scipy-0.13.0/reference/generated/scipy.stats.lognorm.html
# Same result as EXCEL function "=LOGNORM.VERT(x;LT_LN;SG_LN;TRUE)"
Inflow_LN          = np.zeros((300))
Inflow_LN[0:20]    = np.array([1,2,3,4,5,6,7,8,2,3,5,6,7,1,2,4,5,6,7,3])
LN_Stock_Reference_2060 = 5.8693165260386 # calculated independently with Excel

""" Test case with folded normally distributed lifetime for inflow-driven model"""
Time_T_FN  = np.arange(0, 100, 1)
mu    = 20
sg    = 20
lifetime_FoldNorm = {'Type': 'FoldedNormal', 'Mean': np.array([mu]), 'StdDev': np.array([sg])} # call with mu and sigma of curve BEFORE folding,
# dsm makes conversion
Inflow_FN          = np.zeros((100))
Inflow_FN[0:20]    = np.array([1,2,3,4,5,6,7,8,2,3,5,6,7,1,2,4,5,6,7,3])
FN_Stock_Reference_2060 = 6.5004904790008300000
 # calculated independently with Excel

# Inflow for quick declines in stock, with negative entries, no correction of negative inflows
InflowNeg_NoCorr = np.array([0,0,1.003845108798,1.009876170898,1.022946871294,2.052181974193,1.098597211758,2.178632487261,0.2864589383765,
-1.57838035116,0.5823071731028,0.7368688072826,0.8609818446092,0.9303172472582,2.940380920226,0.8868056550348,0.8077321288789,1.74261978268,
1.72284735601,2.770808172213,-2.13996861954,0.9927014953087,1.118058099656,1.211591711403,1.256322506739,1.248773108954,1.197964611415,
1.122904133751,1.047910488113,0.9952110047915])
InflowNeg_WithCorr = np.array([0,0,1.003845108797608,1.009876170898179,1.022946871293978,2.052181974192771,1.09859721175781,2.178632487260613,
0.2864589383765391,0,0.468909567774503,0.600148064601485,0.7138380402455925,0.7922474677487478,2.832866490884595,0.8262270093751778,0.7986846781384966,
1.775290769026188,1.776531605001062,2.822224969997677,0,0.82099326892646,0.9183009967181216,1.007417665416504,1.076751857993922,1.119633273700667,
1.134233884697729,1.123750382153023,1.096483322433279,1.064262457376682])

StockChange_WithCorr = np.array([0,0,1,1,1,2,1,2,0,-2,0,0,0,0,2,0,0,1,1,2,-3,0,0,0,0,0,0,0,0,0])

InitialStockInflowNegNoCorrect = np.array([0.7929476381216127,0.6688896434012339,0.2377146834690121,0.4401462789619187,1.050188897357724,
1.023279749316858,1.00991262429366,1.003845108797608,2.734512232505172,-1.576179715364555,0.5270786116039677,0.6253099746885178,
0.7133245646541703,0.7808737550728377,-2.193761739572536,0.8006372758619281,0.7486059179704083,0.6628542885032971,1.562532671694373,
1.463347346524538,1.385763022717013,1.346937342733784,1.360199865992007,1.431582763055532,0.5522778871344547,0.7074660291917115,
-1.038359636072531,1.901306859994554,1.088965343251684,1.121312374405391])

InitialStockInflowNoNegCorrect = np.array([0.7929476381216127,0.6688896434012339,0.2377146834690121,0.4401462789619187,1.050188897357724,
1.023279749316858,1.00991262429366,1.003845108797608,2.734512232505172,0.431510502230662,0.5391407358051091,0.8522203972396362,
1.568386852314742,0.8767713133566062,0.4696866483782957,0.5372828546582809,0.5715642400086977,0.571388710240865,0.5417776639498378,
0.4910906219746665,0.4275448257791661,2.365060643030991,1.802882915501588,1.755204649624293,0.5284632857607936,0.4360513595181573,
0.3800804006621792,0.755755759275017,0.9453763220611762,1.024955581556197])

InitialStockInflowNegCorrect = np.array([0.7929476381216127,0.6688896434012339,0.2377146834690121,0.4401462789619187,
1.050188897357724,1.023279749316858,1.00991262429366,1.003845108797608,2.734512232505172,0,0.4252614122673855,0.5118790151618888,
0.5969693316522457,0.6739530024236163,0,0.4358419187951329,0.4372674850950349,0.4227905939406414,1.401739961117737,1.38013886282689,
1.370084205085089,1.381499862452395,1.422069169849336,1.495945462462685,0.5985535082458745,0.7251533575497293,0,1.756743048425328,
0.9375744022016721,0.984738434681199])

# Parameters for stock-driven model with initial stock:
FutureStock1  = np.array([2.8,3,3.5,4,5,5])
FutureStock1a = np.array([0,0,0,2.8,3,3.5,4,5,5])
FutureStock2  = np.array([2.8,3,3.5,2.2,2.3,2.5])
FutureStock2a = np.array([0,0,0,2.8,3,3.5,2.2,2.3,2.5])
InitialStock1 = np.array([[0.4,0.2,0.5,0,0,0,0,0,0],[0.1,0.5,1.0,0,0,0,0,0,0]]).transpose()
InitialStock2 = np.array([[0.4,0.2,0.5,0.1,0.25,0.3,0.1,0.1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                          [0.1,0.5,1.0,0.7,0.2,0.3,0.45,0.2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]]).transpose()
TypeSplit1    = np.array([[0.5,0.5,0.1,0.5,0.8,0.5],[0.5,0.5,0.9,0.5,0.2,0.5]]).transpose()
TypeSplit1a   = np.array([[0,0,0,0.5,0.5,0.1,0.5,0.8,0.5],[0,0,0,0.5,0.5,0.9,0.5,0.2,0.5]]).transpose()
TypeSplit2a   = np.array([[0,0,0,0,0,0,0,0,0.5,0.5,0.1,0.5,0.8,0.5,0.5,0.1,0.5,0.8,0.5,0.5,0.1,0.5,0.8,0.5,0.5,0.1,0.5,0.8,0.5,0.33],
                          [0,0,0,0,0,0,0,0,0.5,0.5,0.9,0.5,0.2,0.5,0.5,0.9,0.5,0.2,0.5,0.5,0.9,0.5,0.2,0.5,0.5,0.9,0.5,0.2,0.5,0.67]]).transpose()

TypeSplitStockCheckType1=np.array([0.1520464589202331,0.06727086112098983,0.1559267249296865,0.1326551318262058,0.1923019570573063,
                                   0.06734416832144496,0.4795521109040651,1.459402923878064,0])

TypeSplitStockCheckType2=np.array([0.03801161473005826,0.1681771528024746,0.311853449859373,0.1326551318262058,0.1923019570573063,
                                   0.6060975148930047,0.4795521109040651,0.3648507309695161,0])

TypeSplitInflowCheckType=np.array([[0.6761333517696461,0.1690333379424115],
[0.2558444008518857,0.6396110021297141],
[0.5,1],
[0.3523465342883199,0.3523465342883199],
[0.4135884347844512,0.4135884347844512],
[0.1138340956232945,1.02450686060965],
[0.6134536124575378,0.6134536124575378],
[1.459402923878064,0.3648507309695161],
[0.5136958913485281,0.5136958913485281]])

TypeSplitInitialStockInflowCorrect1StockCheck = np.array([[0,0],
[0,0],
[0.5,1],
[0.3908625698550751,0.7817251397101502],
[0.2957996369747763,0.5915992739495526],
[0.2324798530180466,0.4649597060360933],
[0.1882452627129519,0.3764905254259038],
[0.1559267249296865,0.311853449859373],
[0.1314683082705697,0.2629366165411394]])

TypeSplitInitialStockInflowCorrect2StockCheck = np.array([[0,0],
[0,0],
[0.5,1],
[0.3908625698550751,0.7817251397101502],
[0.2957996369747763,0.5915992739495526],
[0.2324798530180466,0.4649597060360933],
[0.1882452627129519,0.3764905254259038],
[0.1559267249296865,0.311853449859373],
[0.1314683082705697,0.2629366165411394]])

TypeSplitInitialStockInflowCorrect3StockCheck = np.array([[0,0],
[0,0],
[0.5,1],
[0.3908625698550751,0.7817251397101502],
[0.2957996369747763,0.5915992739495526],
[0.2324798530180466,0.4649597060360933],
[0.1493421286476113,0.2986842572952226],
[0.123702603074578,0.247405206149156],
[0.1042988106254021,0.2085976212508043]])

###############################################################################
"""Create Dynamic Stock Models and hand over the pre-defined values."""
# For zero lifetime: border case
myDSM0 = dsm.DynamicStockModel(t=Time_T_FixedLT, i=Inflow_T_FixedLT, lt=lifetime_FixedLT0)

# For fixed LT
myDSM = dsm.DynamicStockModel(t=Time_T_FixedLT, i=Inflow_T_FixedLT, lt=lifetime_FixedLT)

myDSM2 = dsm.DynamicStockModel(t=Time_T_FixedLT, s=Stock_T_FixedLT, lt=lifetime_FixedLT)

myDSMx = dsm.DynamicStockModel(t=Time_T_FixedLT_X, lt=lifetime_FixedLT_X)
TestInflow_X = myDSMx.compute_i_from_s(InitialStock=InitialStock_X)

myDSMxy = dsm.DynamicStockModel(t=Time_T_FixedLT_X, i=TestInflow_X, lt=lifetime_FixedLT_X)

# For zero normally distributed lifetime: border case
myDSM0n = dsm.DynamicStockModel(t=Time_T_FixedLT, i=Inflow_T_FixedLT, lt=lifetime_NormLT0)

# For normally distributed Lt
myDSM3 = dsm.DynamicStockModel(t=Time_T_FixedLT, i=Inflow_T_FixedLT, lt=lifetime_NormLT)

myDSM4 = dsm.DynamicStockModel(t=Time_T_FixedLT, s=Stock_T_NormLT, lt=lifetime_NormLT)
myDSM4a= dsm.DynamicStockModel(t=Time_T_FixedLT, s=Stock_T_NormLT, lt=lifetime_NormLT)

myDSMX = dsm.DynamicStockModel(t=Time_T_FixedLT_XX, lt=lifetime_NormLT_X)
TestInflow_XX = myDSMX.compute_i_from_s(InitialStock=InitialStock_XX)

myDSMXY = dsm.DynamicStockModel(t=Time_T_FixedLT_XX, i=TestInflow_XX, lt=lifetime_NormLT_X)

#For negatative inflow correction 
myDSM_ICF = dsm.DynamicStockModel(t=Time_T_30, s=Stock_SDM_NegInflow_NormLT, lt=lifetime_NormLT8)
[SC1,OC1,I1] = myDSM_ICF.compute_stock_driven_model(NegativeInflowCorrect = False)
myDSM_ICT = dsm.DynamicStockModel(t=Time_T_30, s=Stock_SDM_NegInflow_NormLT, lt=lifetime_NormLT8)
[SC2,OC2,I2] = myDSM_ICT.compute_stock_driven_model(NegativeInflowCorrect = True)
OT2 = myDSM_ICT.compute_outflow_total()
SCh2= myDSM_ICT.compute_stock_change()
Bal2= myDSM_ICT.check_stock_balance()

#For negatative inflow correction with initial stock
#a) Neg inflow, Correction Flag is False:
myDSM_ICFIS = dsm.DynamicStockModel(t=Time_T_30, s=Stock_SDM_NegInflow_NormLTNeg.copy(), lt=lifetime_NormLT8)
[SC1IS,OC1IS,I1IS] = myDSM_ICFIS.compute_stock_driven_model_initialstock(InitialStock = InitialStock_8, SwitchTime = 9, NegativeInflowCorrect = False)
OT1IS = myDSM_ICFIS.compute_outflow_total()
SCh1IS= myDSM_ICFIS.compute_stock_change()
Bal1IS= myDSM_ICFIS.check_stock_balance()
#b) No neg. inflow, correction flag is True:
myDSM_ICTIST = dsm.DynamicStockModel(t=Time_T_30, s=Stock_SDM_PosInflow_NormLTNeg.copy(), lt=lifetime_NormLT8)
[SC3IS,OC3IS,I3IS] = myDSM_ICTIST.compute_stock_driven_model_initialstock(InitialStock = InitialStock_8, SwitchTime = 9, NegativeInflowCorrect = True)
OT3IS = myDSM_ICTIST.compute_outflow_total()
SCh3IS= myDSM_ICTIST.compute_stock_change()
Bal3IS= myDSM_ICTIST.check_stock_balance()
#c) Neg. inflow: Correction flag is True and neg. inflow occurs.
myDSM_ICTIS = dsm.DynamicStockModel(t=Time_T_30, s=Stock_SDM_NegInflow_NormLTNeg.copy(), lt=lifetime_NormLT8)
[SC2IS,OC2IS,I2IS] = myDSM_ICTIS.compute_stock_driven_model_initialstock(InitialStock = InitialStock_8, SwitchTime = 9, NegativeInflowCorrect = True)
OT2IS = myDSM_ICTIS.compute_outflow_total()
SCh2IS= myDSM_ICTIS.compute_stock_change()
Bal2IS= myDSM_ICTIS.check_stock_balance()

# Test compute_stock_driven_model_initialstock:
TestDSM_IntitialStock = dsm.DynamicStockModel(t=Time_T_FixedLT_2, s=FutureStock_2, lt=lifetime_NormLT_2)
Sc_InitialStock_2,Oc_InitialStock_2,I_InitialStock_2 = TestDSM_IntitialStock.compute_stock_driven_model_initialstock(InitialStock = InitialStock_2, SwitchTime = ThisSwitchTime)
OT3 = TestDSM_IntitialStock.compute_outflow_total()
#TestDSM_IntitialStock.compute_stock_total()
SCh3= TestDSM_IntitialStock.compute_stock_change()
Bal3= TestDSM_IntitialStock.check_stock_balance()

# Test stock-driven model with initialstock and type split
TestDSM_IntitialStockTypeSplit = dsm.DynamicStockModel(t=Time_T_FixedLT_2, s=FutureStock1, lt=lifetime_NormLT_3)
SFArrayCombined                = TestDSM_IntitialStockTypeSplit.compute_sf().copy()
[SC_IS_TS,OC_IS_TS,I_IS_TS]    = TestDSM_IntitialStockTypeSplit.compute_stock_driven_model_initialstock_typesplit(FutureStock=FutureStock1,InitialStock=InitialStock1,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombined,np.ones(2)),TypeSplit=TypeSplit1)
OT_IS_TS  = TestDSM_IntitialStockTypeSplit.compute_outflow_total()
SCh_IS_TS = TestDSM_IntitialStockTypeSplit.compute_stock_change()
Bal_IS_TS = TestDSM_IntitialStockTypeSplit.check_stock_balance()

# Test stock-driven model with initialstock and type split and NegativeInflowCorrect, lognormal lifetime
# 1) NegativeInflowCorrect on but does not occur
TestDSM_IntitialStockTypeSplit1       = dsm.DynamicStockModel(t=Time_T_FixedLT_2, s=FutureStock1a.copy(), lt=lifetime_NormLT_3) 
SFArrayCombined                       = TestDSM_IntitialStockTypeSplit1.compute_sf().copy()
[SC_IS_TS1,OC_IS_TS1,I_IS_TS1,Flags1] = TestDSM_IntitialStockTypeSplit1.compute_stock_driven_model_initialstock_typesplit_negativeinflowcorrect(SwitchTime=3,InitialStock=InitialStock1,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombined,np.ones(2)),TypeSplit=TypeSplit1a,NegativeInflowCorrect=True)
OT_IS_TS1                             = TestDSM_IntitialStockTypeSplit1.compute_outflow_total()
SCh_IS_TS1                            = TestDSM_IntitialStockTypeSplit1.compute_stock_change()
Bal_IS_TS1                            = TestDSM_IntitialStockTypeSplit1.check_stock_balance()

# 2) NegativeInflowCorrect on but is not fixed
TestDSM_IntitialStockTypeSplit2       = dsm.DynamicStockModel(t=Time_T_FixedLT_2, s=FutureStock2a.copy(), lt=lifetime_NormLT_3) 
SFArrayCombined                       = TestDSM_IntitialStockTypeSplit2.compute_sf().copy()
[SC_IS_TS2,OC_IS_TS2,I_IS_TS2,Flags2] = TestDSM_IntitialStockTypeSplit2.compute_stock_driven_model_initialstock_typesplit_negativeinflowcorrect(SwitchTime=3,InitialStock=InitialStock1,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombined,np.ones(2)),TypeSplit=TypeSplit1a,NegativeInflowCorrect=False)
OT_IS_TS2                             = TestDSM_IntitialStockTypeSplit2.compute_outflow_total()
SCh_IS_TS2                            = TestDSM_IntitialStockTypeSplit2.compute_stock_change()
Bal_IS_TS2                            = TestDSM_IntitialStockTypeSplit2.check_stock_balance()

# 3) NegativeInflowCorrect on and fixed
TestDSM_IntitialStockTypeSplit3       = dsm.DynamicStockModel(t=Time_T_FixedLT_2, s=FutureStock2a.copy(), lt=lifetime_NormLT_3) 
SFArrayCombined                       = TestDSM_IntitialStockTypeSplit3.compute_sf().copy()
[SC_IS_TS3,OC_IS_TS3,I_IS_TS3,Flags3] = TestDSM_IntitialStockTypeSplit3.compute_stock_driven_model_initialstock_typesplit_negativeinflowcorrect(SwitchTime=3,InitialStock=InitialStock1,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombined,np.ones(2)),TypeSplit=TypeSplit1a,NegativeInflowCorrect=True)
OT_IS_TS3                             = TestDSM_IntitialStockTypeSplit3.compute_outflow_total()
SCh_IS_TS3                            = TestDSM_IntitialStockTypeSplit3.compute_stock_change()
Bal_IS_TS3                            = TestDSM_IntitialStockTypeSplit3.check_stock_balance()

# Test stock-driven model with initialstock and type split and NegativeInflowCorrect, normal lifetime
# 1) NegativeInflowCorrect on but does not occur
TestDSM_IntitialStockTypeSplit1a          = dsm.DynamicStockModel(t=Time_T_FixedLT_3, s=Stock_SDM_PosInflow_NormLTNeg.copy(), lt=lifetime_NormLT_4) 
SFArrayCombineda                          = TestDSM_IntitialStockTypeSplit1a.compute_sf().copy()
[SC_IS_TS1a,OC_IS_TS1a,I_IS_TS1a,Flags1a] = TestDSM_IntitialStockTypeSplit1a.compute_stock_driven_model_initialstock_typesplit_negativeinflowcorrect(SwitchTime=8,InitialStock=InitialStock2,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombineda,np.ones(2)),TypeSplit=TypeSplit2a,NegativeInflowCorrect=True)
OT_IS_TS1a                                = TestDSM_IntitialStockTypeSplit1a.compute_outflow_total()
SCh_IS_TS1a                               = TestDSM_IntitialStockTypeSplit1a.compute_stock_change()
Bal_IS_TS1a                               = TestDSM_IntitialStockTypeSplit1a.check_stock_balance()

# 2) NegativeInflowCorrect on but is not fixed
TestDSM_IntitialStockTypeSplit2a          = dsm.DynamicStockModel(t=Time_T_FixedLT_3, s=Stock_SDM_NegInflow_NormLTNeg.copy(), lt=lifetime_NormLT_4) 
SFArrayCombineda                          = TestDSM_IntitialStockTypeSplit2a.compute_sf().copy()
[SC_IS_TS2a,OC_IS_TS2a,I_IS_TS2a,Flags2a] = TestDSM_IntitialStockTypeSplit2a.compute_stock_driven_model_initialstock_typesplit_negativeinflowcorrect(SwitchTime=8,InitialStock=InitialStock2,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombineda,np.ones(2)),TypeSplit=TypeSplit2a,NegativeInflowCorrect=False)
OT_IS_TS2a                                = TestDSM_IntitialStockTypeSplit2a.compute_outflow_total()
SCh_IS_TS2a                               = TestDSM_IntitialStockTypeSplit2a.compute_stock_change()
Bal_IS_TS2a                               = TestDSM_IntitialStockTypeSplit2a.check_stock_balance()

# 3) NegativeInflowCorrect on and fixed
TestDSM_IntitialStockTypeSplit3a          = dsm.DynamicStockModel(t=Time_T_FixedLT_3, s=Stock_SDM_NegInflow_NormLTNeg.copy(), lt=lifetime_NormLT_4) 
SFArrayCombineda                          = TestDSM_IntitialStockTypeSplit3a.compute_sf().copy()
[SC_IS_TS3a,OC_IS_TS3a,I_IS_TS3a,Flags3a] = TestDSM_IntitialStockTypeSplit3a.compute_stock_driven_model_initialstock_typesplit_negativeinflowcorrect(SwitchTime=8,InitialStock=InitialStock2,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombineda,np.ones(2)),TypeSplit=TypeSplit2a,NegativeInflowCorrect=True)
OT_IS_TS3a                                = TestDSM_IntitialStockTypeSplit3a.compute_outflow_total()
SCh_IS_TS3a                               = TestDSM_IntitialStockTypeSplit3a.compute_stock_change()
Bal_IS_TS3a                               = TestDSM_IntitialStockTypeSplit3a.check_stock_balance()

# Compute stock back from inflow
TestDSM_IntitialStock_Verify = dsm.DynamicStockModel(t=Time_T_FixedLT_2, i=I_InitialStock_2, lt=lifetime_NormLT_2)
Sc_Stock_2 = TestDSM_IntitialStock_Verify.compute_s_c_inflow_driven()
Sc_Stock_2_Sum = Sc_Stock_2.sum(axis =1)
Sc_Stock_Sum   = TestDSM_IntitialStock_Verify.compute_stock_total()
Sc_Outflow_t_c = TestDSM_IntitialStock_Verify.compute_o_c_from_s_c()

# For Weibull-distributed Lt
myDSMWB1 = dsm.DynamicStockModel(t=Time_T_FixedLT, i=Inflow_T_FixedLT, lt=lifetime_WeibullLT)

myDSMWB2 = dsm.DynamicStockModel(t=Time_T_FixedLT, s=Stock_T_WeibullLT, lt=lifetime_WeibullLT)

myDSMWB3 = dsm.DynamicStockModel(t=Time_T_FixedLT_XX, lt=lifetime_WeibullLT)
TestInflow_WB = myDSMWB3.compute_i_from_s(InitialStock=InitialStock_XX)

myDSMWB4 = dsm.DynamicStockModel(t=Time_T_FixedLT_XX, i=TestInflow_WB, lt=lifetime_WeibullLT)
# Compute full stock model in correct order

# Test lognormal lifetime dist.
myDSM_LN = dsm.DynamicStockModel(t=Time_T_LN, i=Inflow_LN, lt=lifetime_LogNorm)
LN_Stock = myDSM_LN.compute_s_c_inflow_driven()
LN_Stock_60 = LN_Stock.sum(axis =1)

# Test folded normal lifetime dist.
myDSM_FN = dsm.DynamicStockModel(t=Time_T_FN, i=Inflow_FN, lt=lifetime_FoldNorm)
FN_Stock = myDSM_FN.compute_s_c_inflow_driven()
FN_Stock_60 = FN_Stock.sum(axis =1)


# Test that ouflow computed with pdf from compute_outflow_pdf() yields same result as the one calculated directly from the sf via compute_s_c_inflow_driven() and compute_o_c_from_s_c()
lt_pdf   = {'Type': 'Normal', 'Mean': np.array([8]), 'StdDev': np.array([3])}
ifl_pdf  = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 1, 2, 3, 4, 5]

DSM_pdf = dsm.DynamicStockModel(t=np.arange(0, 25, 1),i=ifl_pdf, lt=lt_pdf)
SF_pdf  = DSM_pdf.compute_sf()
PDF_pdf = DSM_pdf.compute_outflow_pdf()
SC_pdf  = DSM_pdf.compute_s_c_inflow_driven()
OC_pdf  = DSM_pdf.compute_o_c_from_s_c()
OC_alt  = np.einsum('c,tc->tc',np.array(ifl_pdf),PDF_pdf)
Bal_pdf = (np.abs(OC_pdf-OC_alt)).sum()
###############################################################################
"""Unit Test Class"""


class KnownResultsTestCase(unittest.TestCase):

    def test_inflow_driven_model_fixedLifetime_0(self):
        """Test Inflow Driven Model with Fixed product lifetime of 0."""
        np.testing.assert_array_equal(myDSM0.compute_s_c_inflow_driven(), np.zeros(Stock_TC_FixedLT.shape))
        np.testing.assert_array_equal(myDSM0.compute_stock_total(), np.zeros((Stock_TC_FixedLT.shape[0])))
        np.testing.assert_array_equal(myDSM0.compute_stock_change(), np.zeros((Stock_TC_FixedLT.shape[0])))
        np.testing.assert_array_equal(myDSM0.compute_outflow_mb(), Inflow_T_FixedLT)
        np.testing.assert_array_equal(myDSM0.check_stock_balance(), Bal.transpose())
    
    
    def test_inflow_driven_model_fixedLifetime(self):
        """Test Inflow Driven Model with Fixed product lifetime."""
        np.testing.assert_array_equal(myDSM.compute_s_c_inflow_driven(), Stock_TC_FixedLT)
        np.testing.assert_array_equal(myDSM.compute_stock_total(),Stock_T_FixedLT)
        np.testing.assert_array_equal(myDSM.compute_o_c_from_s_c(), Outflow_TC_FixedLT)
        np.testing.assert_array_equal(myDSM.compute_outflow_total(), Outflow_T_FixedLT)
        np.testing.assert_array_equal(myDSM.compute_stock_change(), StockChange_T_FixedLT)
        np.testing.assert_array_equal(myDSM.check_stock_balance(), Bal.transpose())

    def test_stock_driven_model_fixedLifetime(self):
        """Test Stock Driven Model with Fixed product lifetime."""
        np.testing.assert_array_equal(myDSM2.compute_stock_driven_model()[0], Stock_TC_FixedLT)
        np.testing.assert_array_equal(myDSM2.compute_stock_driven_model()[1], Outflow_TC_FixedLT)
        np.testing.assert_array_equal(myDSM2.compute_stock_driven_model()[2], Inflow_T_FixedLT)
        np.testing.assert_array_equal(myDSM2.compute_outflow_total(), Outflow_T_FixedLT)
        np.testing.assert_array_equal(myDSM2.compute_stock_change(), StockChange_T_FixedLT)
        np.testing.assert_array_equal(myDSM2.check_stock_balance(), Bal.transpose())
        
    def test_inflow_driven_model_normallyDistrLifetime_0(self):
        """Test Inflow Driven Model with Fixed product lifetime of 0."""
        np.testing.assert_array_equal(myDSM0n.compute_s_c_inflow_driven(), np.zeros(Stock_TC_FixedLT.shape))
        np.testing.assert_array_equal(myDSM0n.compute_stock_total(), np.zeros((Stock_TC_FixedLT.shape[0])))
        np.testing.assert_array_equal(myDSM0n.compute_stock_change(), np.zeros((Stock_TC_FixedLT.shape[0])))
        np.testing.assert_array_equal(myDSM0n.compute_outflow_mb(), Inflow_T_FixedLT)
        np.testing.assert_array_equal(myDSM0n.check_stock_balance(), Bal.transpose())
            
    def test_inflow_driven_model_normallyDistLifetime(self):
        """Test Inflow Driven Model with normally distributed product lifetime."""
        np.testing.assert_array_almost_equal(myDSM3.compute_s_c_inflow_driven(), Stock_TC_NormLT, 8)
        np.testing.assert_array_almost_equal(myDSM3.compute_stock_total(), Stock_T_NormLT, 8)
        np.testing.assert_array_almost_equal(myDSM3.compute_o_c_from_s_c(), Outflow_TC_NormLT, 8)
        np.testing.assert_array_almost_equal(myDSM3.compute_outflow_total(), Outflow_T_NormLT, 8)
        np.testing.assert_array_almost_equal(myDSM3.compute_stock_change(), StockChange_T_NormLT, 8)
        np.testing.assert_array_almost_equal(myDSM3.check_stock_balance(), Bal.transpose(), 12)

    def test_stock_driven_model_normallyDistLifetime(self):
        """Test Stock Driven Model with normally distributed product lifetime."""
        np.testing.assert_array_almost_equal(
            myDSM4.compute_stock_driven_model()[0], Stock_TC_NormLT, 8)
        np.testing.assert_array_almost_equal(
            myDSM4.compute_stock_driven_model()[1], Outflow_TC_NormLT, 8)
        np.testing.assert_array_almost_equal(
            myDSM4.compute_stock_driven_model()[2], Inflow_T_FixedLT, 8)
        np.testing.assert_array_almost_equal(myDSM4.compute_outflow_total(), Outflow_T_NormLT, 8)
        np.testing.assert_array_almost_equal(
            myDSM4.compute_stock_change(), StockChange_T_NormLT, 8)
        np.testing.assert_array_almost_equal(myDSM4.check_stock_balance(), Bal.transpose(), 12)
        
    def test_stock_driven_model_normallyDistLifetime_NegInflowFlagTrue(self):
        """Test Stock Driven Model with normally distributed product lifetime.
        Set the NegativeInflowCorrect flag as True but use a test case without negative inflows."""
        np.testing.assert_array_almost_equal(
            myDSM4a.compute_stock_driven_model(NegativeInflowCorrect = True)[0], Stock_TC_NormLT, 8)
        np.testing.assert_array_almost_equal(
            myDSM4a.compute_stock_driven_model(NegativeInflowCorrect = True)[1], Outflow_TC_NormLT, 8)
        np.testing.assert_array_almost_equal(
            myDSM4a.compute_stock_driven_model(NegativeInflowCorrect = True)[2], Inflow_T_FixedLT, 8)
        np.testing.assert_array_almost_equal(myDSM4a.compute_outflow_total(), Outflow_T_NormLT, 8)
        np.testing.assert_array_almost_equal(
            myDSM4a.compute_stock_change(), StockChange_T_NormLT, 8)
        np.testing.assert_array_almost_equal(myDSM4a.check_stock_balance(), Bal.transpose(), 12)        
        
    def test_stock_driven_model_normallyDistLifetime_NegInflow(self):
        """Test Stock Driven Model with normally distributed product lifetime, with negative inflow."""
        np.testing.assert_array_almost_equal(
            myDSM_ICF.compute_stock_driven_model(NegativeInflowCorrect = False)[2], InflowNeg_NoCorr, 12)
        np.testing.assert_array_almost_equal(
            myDSM_ICT.compute_stock_driven_model(NegativeInflowCorrect = True)[2], InflowNeg_WithCorr, 12)
        np.testing.assert_array_almost_equal(
            myDSM_ICT.compute_stock_change(), StockChange_WithCorr, 12)
        np.testing.assert_array_almost_equal(myDSM_ICT.check_stock_balance(), Bal30.transpose(), 12)   
        
    def test_stock_driven_model_Initialstock_NegInflow(self):
        """Test Stock Driven Model with normally distributed product lifetime, initial stock, with negative inflow and no negative inflow correction."""
        np.testing.assert_array_almost_equal(myDSM_ICFIS.check_stock_balance(), Bal30.transpose(), 12) 
        np.testing.assert_array_almost_equal(myDSM_ICFIS.compute_stock_driven_model_initialstock(InitialStock = InitialStock_8, SwitchTime = 9, NegativeInflowCorrect = False)[2], InitialStockInflowNegNoCorrect, 12) 
        
    def test_stock_driven_model_Initialstock_NegInflowNotUsed(self):
        """Test Stock Driven Model with normally distributed product lifetime, initial stock, with no negative inflow and no negative inflow correction."""
        np.testing.assert_array_almost_equal(myDSM_ICTIST.check_stock_balance(), Bal30.transpose(), 12) 
        np.testing.assert_array_almost_equal(myDSM_ICTIST.compute_stock_driven_model_initialstock(InitialStock = InitialStock_8, SwitchTime = 9, NegativeInflowCorrect = False)[2], InitialStockInflowNoNegCorrect, 12) 

    def test_stock_driven_model_Initialstock_NegInflowFixed(self):
        """Test Stock Driven Model with normally distributed product lifetime, initial stock, with negative inflow and negative inflow correction."""
        np.testing.assert_array_almost_equal(myDSM_ICTIS.check_stock_balance(), Bal30.transpose(), 12) 
        np.testing.assert_array_almost_equal(myDSM_ICTIS.compute_stock_driven_model_initialstock(InitialStock = InitialStock_8, SwitchTime = 9, NegativeInflowCorrect = True)[2], InitialStockInflowNegCorrect, 12) 

    def test_stock_driven_model_Initialstock_NegInflow_TypeSplit_NotUsed(self):
        """Test Stock Driven Model with lognormally distributed product lifetime, initial stock, typesplit, with negative inflow correction but no negative inflow occurring."""
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit1.check_stock_balance(), Bal9.transpose(), 12) 
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit1.compute_stock_driven_model_initialstock_typesplit_negativeinflowcorrect(SwitchTime=3,InitialStock=InitialStock1,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombined,np.ones(2)),TypeSplit=TypeSplit1a,NegativeInflowCorrect=True)[0][:,2,:], TypeSplitInitialStockInflowCorrect1StockCheck, 12) 
     
    def test_stock_driven_model_Initialstock_NegInflow_TypeSplit_NoCorr(self):
        """Test Stock Driven Model with lognormally distributed product lifetime, initial stock, typesplit, with negative inflow correction of but negative inflow occurring."""
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit2.check_stock_balance(), Bal9.transpose(), 12) 
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit2.compute_stock_driven_model_initialstock_typesplit_negativeinflowcorrect(SwitchTime=3,InitialStock=InitialStock1,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombined,np.ones(2)),TypeSplit=TypeSplit1a,NegativeInflowCorrect=False)[0][:,2,:], TypeSplitInitialStockInflowCorrect2StockCheck, 12) 
        
    def test_stock_driven_model_Initialstock_NegInflow_TypeSplit_Corr(self):
        """Test Stock Driven Model with lognormally distributed product lifetime, initial stock, typesplit, with negative inflow correction on and negative inflow occurring and corrected."""
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit3.check_stock_balance(), Bal9.transpose(), 12) 
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit3.compute_stock_driven_model_initialstock_typesplit_negativeinflowcorrect(SwitchTime=3,InitialStock=InitialStock1,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombined,np.ones(2)),TypeSplit=TypeSplit1a,NegativeInflowCorrect=True)[0][:,2,:], TypeSplitInitialStockInflowCorrect3StockCheck, 12) 

    def test_stock_driven_model_Initialstock_NegInflow_TypeSplit_NotUsed_NormalLT(self):
        """Test Stock Driven Model with normally distributed product lifetime, initial stock, typesplit, with negative inflow correction but no negative inflow occurring."""
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit1a.check_stock_balance(), Bal30.transpose(), 12) 
     
    def test_stock_driven_model_Initialstock_NegInflow_TypeSplit_NoCorr_NormalLT(self):
        """Test Stock Driven Model with normally distributed product lifetime, initial stock, typesplit, with negative inflow correction of but negative inflow occurring."""
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit2a.check_stock_balance(), Bal30.transpose(), 12) 
        
    def test_stock_driven_model_Initialstock_NegInflow_TypeSplit_Corr_NormalLT(self):
        """Test Stock Driven Model with normally distributed product lifetime, initial stock, typesplit, with negative inflow correction on and negative inflow occurring and corrected."""
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit3a.check_stock_balance(), Bal30.transpose(), 12) 

    def test_stock_driven_model_Initialstock_TypeSplit(self):
        """Test Stock Driven Model with lognormally distributed product lifetime, initial stock, and type split."""
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit.check_stock_balance()[1::], np.zeros((5)), 12) 
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit.compute_stock_driven_model_initialstock_typesplit(FutureStock=FutureStock1,InitialStock=InitialStock1,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombined,np.ones(2)),TypeSplit=TypeSplit1)[0][4,:,0], TypeSplitStockCheckType1, 12) 
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit.compute_stock_driven_model_initialstock_typesplit(FutureStock=FutureStock1,InitialStock=InitialStock1,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombined,np.ones(2)),TypeSplit=TypeSplit1)[0][4,:,1], TypeSplitStockCheckType2, 12) 
        np.testing.assert_array_almost_equal(TestDSM_IntitialStockTypeSplit.compute_stock_driven_model_initialstock_typesplit(FutureStock=FutureStock1,InitialStock=InitialStock1,SFArrayCombined=np.einsum('tc,g->tcg',SFArrayCombined,np.ones(2)),TypeSplit=TypeSplit1)[2],        TypeSplitInflowCheckType, 12) 

    def test_inflow_driven_model_WeibullDistLifetime(self):
        """Test Inflow Driven Model with Weibull-distributed product lifetime."""
        np.testing.assert_array_almost_equal(
            myDSMWB1.compute_s_c_inflow_driven(), Stock_TC_WeibullLT, 9)
        np.testing.assert_array_almost_equal(myDSMWB1.compute_stock_total(), Stock_T_WeibullLT, 8)
        np.testing.assert_array_almost_equal(myDSMWB1.compute_o_c_from_s_c(), Outflow_TC_WeibullLT, 9)
        np.testing.assert_array_almost_equal(myDSMWB1.compute_outflow_total(), Outflow_T_WeibullLT, 9)
        np.testing.assert_array_almost_equal(
            myDSMWB1.compute_stock_change(), StockChange_T_WeibullLT, 9)
        np.testing.assert_array_almost_equal(myDSMWB1.check_stock_balance(), Bal.transpose(), 12)

    def test_stock_driven_model_WeibullDistLifetime(self):
        """Test Stock Driven Model with Weibull-distributed product lifetime."""
        np.testing.assert_array_almost_equal(
            myDSMWB1.compute_stock_driven_model()[0], Stock_TC_WeibullLT, 8)
        np.testing.assert_array_almost_equal(
            myDSMWB1.compute_stock_driven_model()[1], Outflow_TC_WeibullLT, 8)
        np.testing.assert_array_almost_equal(
            myDSMWB1.compute_stock_driven_model()[2], Inflow_T_FixedLT, 8)
        np.testing.assert_array_almost_equal(myDSMWB1.compute_outflow_total(), Outflow_T_WeibullLT, 9)
        np.testing.assert_array_almost_equal(
            myDSMWB1.compute_stock_change(), StockChange_T_WeibullLT, 8)
        np.testing.assert_array_almost_equal(myDSMWB1.check_stock_balance(), Bal.transpose(), 12)


    def test_inflow_from_stock_fixedLifetime(self):
        """Test computation of inflow from stock with Fixed product lifetime."""
        np.testing.assert_array_equal(TestInflow_X, Inflow_X)
        np.testing.assert_array_equal(myDSMxy.compute_s_c_inflow_driven()[-1, :], InitialStock_X)

    def test_inflow_from_stock_normallyDistLifetime(self):
        """Test computation of inflow from stock with normally distributed product lifetime."""
        np.testing.assert_array_almost_equal(TestInflow_XX, Inflow_XX, 8)
        np.testing.assert_array_almost_equal(myDSMXY.compute_s_c_inflow_driven()[-1, :], InitialStock_XX, 9)
            
    def test_inflow_from_stock_WeibullDistLifetime(self):
        """Test computation of inflow from stock with Weibull-distributed product lifetime."""
        np.testing.assert_array_almost_equal(TestInflow_WB, Inflow_WB, 6) 
        np.testing.assert_array_almost_equal(myDSMWB4.compute_s_c_inflow_driven()[-1, :], InitialStock_WB, 8)     
        
    def test_compute_stock_driven_model_initialstock(self):
        """Test stock-driven model with initial stock given."""
        np.testing.assert_array_almost_equal(I_InitialStock_2, I_InitialStock_2_Ref, 8) 
        np.testing.assert_array_almost_equal(Sc_InitialStock_2, Sc_InitialStock_2_Ref, 8)
        np.testing.assert_array_almost_equal(Sc_InitialStock_2.sum(axis =1), Sc_InitialStock_2_Ref_Sum, 8)
        np.testing.assert_array_almost_equal(Oc_InitialStock_2, Oc_InitialStock_2_Ref, 8) 

    def test_stock_from_inflow_LogNormalDistLifetime(self):
        """Test computation of stock from inflow for lognormally distributed product lifetime."""
        np.testing.assert_array_almost_equal(LN_Stock_Reference_2060, LN_Stock_60[60], 9) 

    def test_stock_from_inflow_FoldedNormalDistLifetime(self):
        """Test computation of stock from inflow for folded normally distributed product lifetime."""
        np.testing.assert_array_almost_equal(FN_Stock_Reference_2060, FN_Stock_60[60], 12) 

    def test_compute_outflow_pdf(self):
        """Test wether outflow by cohort is the same if computed with pdf from compute_outflow_pdf vs. the standard routines."""
        np.testing.assert_array_almost_equal(OC_pdf, OC_alt, 14) 
        np.testing.assert_array_almost_equal(Bal_pdf, 0, 12) 


    if __name__ == '__main__':
        unittest.main()




import os
import shutil
import sys
import re
from typing import List, Dict, Any, Union, Optional
import numpy as np
import pandas as pd
from IPython import get_ipython
from .datastructures import Scenario, Flow
import aiphoria.lib.odym.modules.ODYM_Classes as msc

global_path_to_output_dir: Union[str, None] = None


def show_model_parameters(model_params: Dict[str, Any]):
    """
    Show model parameters (= prints the model parameters to console).

    :param model_params: Dictionary of model parameters
    """
    print("Using following parameters for running the model:")
    max_param_len = max([len(name) for name in model_params])
    for param_name, param_value in model_params.items():
        print("\t{:{}} = {}".format(param_name, max_param_len, param_value))


def show_exception_errors(exception: Exception, msg: str = ""):
    """
    Show Exception errors.

    :param msg: Message about exception.
    :param exception: Exception
    """

    if msg:
        print(msg)

    errors = exception.args[0]
    if type(errors) is str:
        print(errors)

    if type(errors) is list:
        for error in errors:
            print("\t{}".format(error))


def setup_current_working_directory():
    """
    Setup current working directory.
    """
    if get_ipython() is None:
        # Running in terminal
        os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
    else:
        # Running Notebook, cwd is already set
        pass


def create_output_directory(output_dir_path: str):
    """
    Create output directory.
    Deletes existing directory.
    """

    # If exists then delete existing directory and create new
    if os.path.exists(output_dir_path):
        try:
            shutil.rmtree(output_dir_path)
        except Exception as ex:
            raise ex

    # Create output directory and directories for all scenarios
    os.makedirs(output_dir_path)


def setup_odym_directories():
    """
    Setup ODYM directories. Appends to path to ODYM files to sys.path.
    """
    sys.path.insert(0, os.path.join(
        os.getcwd(), '.', 'lib', 'odym', 'modules'))


def setup_scenario_output_directories(output_dir_path: str, scenario_names: List[str]) -> Dict[str, str]:
    """
    Setup scenario output directories. Deletes existing scenario directory if exists.
    Returns Dictionary of scenario name to scenario directory (absolute path)

    :param output_dir_name: Output directory name
    :param scenario_names: List of scenario names
    :return: Dictionary [str, str]: Scenario name to scenario directory (absolute path)
    """

    scenario_name_to_abs_scenario_output_path = {}
    for scenario_name in scenario_names:
        abs_scenario_output_path = os.path.join(output_dir_path, scenario_name)
        try:
            shutil.rmtree(abs_scenario_output_path)
        except FileNotFoundError as ex:
            pass
        except Exception as ex:
            raise ex

        os.makedirs(abs_scenario_output_path)
        scenario_name_to_abs_scenario_output_path[scenario_name] = abs_scenario_output_path

    return scenario_name_to_abs_scenario_output_path


def build_mfa_system_for_scenario(scenario: Scenario):
    """
    Build MFA system for scenario.

    :param scenario: Scenario-object
    :param progress_bar: Progress bar instance (optional)
    :return: ODYM MFASystem
    """

    # Track solid wood equivalent and carbon.
    # Dictionary of classifications enters the index table defined for the system.
    # The index table lists all aspects needed and assigns a classification and index letter to each aspect.
    scenario_data = scenario.scenario_data
    flow_solver = scenario.flow_solver
    years = scenario_data.years

    # Baseline value and unit name: e.g. name = "Solid wood equivalent and unit = "Mm3"
    baseline_value_name = scenario_data.baseline_value_name
    baseline_unit_name = scenario_data.baseline_unit_name

    # Indicator names are used as Elements in ODYM MFASystem
    indicator_names = flow_solver.get_indicator_names()

    model_time_start = scenario_data.start_year
    model_time_end = scenario_data.end_year
    model_elements = [baseline_value_name] + indicator_names
    model_years = years

    model_classifications = {
        'Time': msc.Classification(Name='Time', Dimension='Time', ID=1, Items=model_years),
        'Cohort': msc.Classification(Name='Age-cohort', Dimension='Time', ID=2, Items=model_years),
        'Element': msc.Classification(Name='Elements', Dimension='Element', ID=3, Items=model_elements),
    }

    index_table = pd.DataFrame({'Aspect': ['Time', 'Age-cohort', 'Element'],  # 'Time' and 'Element' must be present!
                                'Description': ['Model aspect "time"', 'Model aspect "age-cohort"', 'Model aspect "Element"'],
                                # 'Time' and 'Element' are also dimensions
                                'Dimension': ['Time', 'Time', 'Element'],
                                'Classification': [model_classifications[Aspect] for Aspect in ['Time', 'Cohort', 'Element']],
                                # Unique one letter (upper or lower case) indices to be used later for calculations.
                                'IndexLetter': ['t', 'c', 'e']})

    # Default indexing of IndexTable, other indices are produced on the fly
    index_table.set_index('Aspect', inplace=True)

    # ****************************
    # Initialize ODYM MFA system *
    # ****************************
    flow_solver = scenario.flow_solver
    mfa_system = msc.MFAsystem(Name='MFA System',
                               Geogr_Scope='Not defined',
                               Unit=baseline_unit_name,
                               ProcessList=[], FlowDict={}, StockDict={}, ParameterDict={},
                               Time_Start=model_time_start, Time_End=model_time_end, IndexTable=index_table,
                               Elements=index_table.loc['Element'].Classification.Items)

    # Get inflow values to stock
    year_index_to_year = dict(enumerate(model_years))
    unique_processes = flow_solver.get_unique_processes()
    unique_flows = flow_solver.get_unique_flows()

    # Create ODYM objects
    # print("Building ODYM processes...")
    odym_processes = []
    process_id_to_index = {}
    for process_id, process in unique_processes.items():
        process_index = len(odym_processes)
        process_id_to_index[process_id] = process_index
        new_process = msc.Process(ID=process_index, Name=process.name)
        odym_processes.append(new_process)

    # print("Building ODYM flows...")
    odym_flows = {}
    for flow_id, flow in unique_flows.items():
        source_process_index = process_id_to_index[flow.source_process_id]
        target_process_index = process_id_to_index[flow.target_process_id]
        new_flow = msc.Flow(ID=flow.id, P_Start=source_process_index,
                            P_End=target_process_index, Indices='t,e', Values=None)
        odym_flows[flow.id] = new_flow

    # print("Building ODYM stocks...")
    odym_stocks = {}
    for stock in flow_solver.get_all_stocks():
        process_index = process_id_to_index[stock.id]
        new_stock = msc.Stock(ID=stock.id, Name=stock.name,
                              P_Res=process_index, Indices='t,e', Type=0, Values=None)
        odym_stocks[stock.id] = new_stock

    mfa_system.ProcessList = odym_processes
    mfa_system.FlowDict = odym_flows
    mfa_system.StockDict = odym_stocks
    mfa_system.Initialize_FlowValues()
    mfa_system.Initialize_StockValues()
    mfa_system.Consistency_Check()

    # Update ODYM flow values from flow values DataFrame
    for flow_id, flow in mfa_system.FlowDict.items():
        for year_index, value in enumerate(flow.Values):
            # Skip to next year if FlowSolver does not have data for this year
            # This is possible because ODYM flow and stock values are already initialized to 0.0
            flow_has_data_for_year = flow_solver.has_flow(
                year=year_index_to_year[year_index], flow_id=flow_id)
            if not flow_has_data_for_year:
                continue

            # NOTE: Virtual flows use default value defined in Flow for carbon content (now 1.0).
            # Get all evaluated values and set those to ODYM flow
            solved_flow = flow_solver.get_flow(
                year=year_index_to_year[year_index], flow_id=flow_id)

            # If flow is not present this year, use 0.0 value
            if not isinstance(solved_flow, Flow):
                for index, indicator_name in enumerate(indicator_names):
                    flow.Values[year_index, index] = 0.0
                continue

            for index, evaluated_value in enumerate(solved_flow.get_all_evaluated_values()):
                flow.Values[year_index, index] = evaluated_value

    # Process stocks (fill with data)
    for stock_id, stock in odym_stocks.items():
        # Baseline DSM
        # These are needed for triggering DSM calculations
        baseline_dsm = flow_solver.get_baseline_dynamic_stocks()[stock_id]
        baseline_stock_by_cohort = baseline_dsm.compute_s_c_inflow_driven()
        baseline_outflow_by_cohort = baseline_dsm.compute_o_c_from_s_c()
        baseline_stock_total = baseline_dsm.compute_stock_total()
        baseline_stock_change = baseline_dsm.compute_stock_change()
        stock.Values[:, 0] = baseline_stock_change

        # Indicator DSMs
        for indicator_index, indicator_name in enumerate(indicator_names):
            # These are needed for triggering DSM calculations
            indicator_dsm = flow_solver.get_indicator_dynamic_stocks()[
                stock_id][indicator_name]
            indicator_stock_by_cohort = indicator_dsm.compute_s_c_inflow_driven()
            indicator_outflow_by_cohort = indicator_dsm.compute_o_c_from_s_c()
            indicator_stock_total = indicator_dsm.compute_stock_total()
            indicator_stock_change = indicator_dsm.compute_stock_change()

            # Baseline values are at index 0 so offset by +1
            stock.Values[:, indicator_index + 1] = indicator_stock_change

    return mfa_system


def calculate_scenario_mass_balance(mfa_system: msc.MFAsystem) -> pd.DataFrame:
    """
    Get scenario mass balance difference per year.

    :param mfa_system: MFASystem-object
    :param model_years: List of years
    :return: DataFrame
    """

    mb = mfa_system.MassBalance()
    # print("Mass balance result shape: {}".format(mb.shape))
    df_mass_balance = pd.DataFrame(
        columns=["Year", "Process 0", "Rest", "Difference"])
    for year_index, year in enumerate(mfa_system.Time_L):
        # Calculate mass balance using the first element in MFA system (= base element)
        # Negative value in process 0 means that process 0 has no inflows so this mass
        # is coming from outside system boundaries
        p0 = np.sum(mb[year_index][0][0])
        rest = np.sum(mb[year_index][1:, 0])
        abs_diff = abs(p0) - abs(rest)
        df_mass_balance.loc[year_index] = np.array([year, p0, rest, abs_diff])
    df_mass_balance = df_mass_balance.astype({"Year": "int32"})
    # df_mass_balance.set_index(["Year"], inplace=True)
    return df_mass_balance


def shorten_sheet_name(name, max_length=31):
    """Shorten and sanitize Excel sheet names to comply with the 31-character limit."""
    # Sanitize name
    sanitized_name = re.sub(r'[:/\\]', '_', name)

    # Split by underscores to maintain structure
    parts = sanitized_name.split('_')

    # Identify fixed parts: these are short words (<= 3 chars) or common connectors
    fixed_parts = {i: p for i, p in enumerate(
        parts) if len(p) <= 3 or not p.isalnum()}
    variable_parts = [p for i, p in enumerate(parts) if i not in fixed_parts]

    # Determine available length for variable parts (accounting for underscores and fixed parts)
    num_underscores = len(parts) - 1
    fixed_length = sum(len(p) for p in fixed_parts.values())
    available_length = max_length - num_underscores - fixed_length

    # Reduce variable parts proportionally
    total_variable_length = sum(len(p) for p in variable_parts)
    if total_variable_length <= available_length:
        return sanitized_name[:max_length]  # Trim if needed

    reduced_parts = []
    for part in variable_parts:
        reduced_length = max(
            1, len(part) * available_length // total_variable_length)
        reduced_parts.append(part[:reduced_length])

    # Reconstruct the name, preserving fixed parts
    final_parts = []
    variable_index = 0
    for i in range(len(parts)):
        if i in fixed_parts:
            final_parts.append(fixed_parts[i])
        else:
            final_parts.append(reduced_parts[variable_index])
            variable_index += 1

    # Ensure final length
    shortened_name = "_".join(final_parts)
    return shortened_name[:max_length]


def get_output_directory() -> str:
    """
    Get output path directory.
    NOTE: Must be using the set_output_directory(), otherwise raises Exception
    """

    if global_path_to_output_dir is None:
        raise Exception("Output path not set")

    return global_path_to_output_dir


def set_output_directory(path_to_output_dir: str) -> None:
    """
        Set output directory for results.
        Path is converted to absolute path (e.g. .., . and ~ are converted to
        actual directory names).

        :param path_to_output_dir:  Path to directory

    """
    # NOTE: os.path.realpath follows symlinks where os.path.abspath does not
    # NOTE: ~ is not expanded to user home directory in Linux
    global_path_to_output_dir = os.path.realpath(path_to_output_dir)


def make_output_directory(path_to_output: str = "", remove_existing: bool = True) -> None:
    """
        Make output directory.
        If path to output directory is not set then creates directory "output"
        by default to the directory of the running script.

        :param remove_existing: If True then removes existing directory.
    """
    # TODO: Implement creating output directory

    os.makedirs(exist_ok=True)



import json
import os
import webbrowser
from typing import Dict, Any
from .datastructures import ScenarioData
from importlib.resources import files


class NetworkGraph(object):
    def __init__(self):
        self._echarts = ""
        self._visualizer = ""
        self._html = ""

        # Load ECharts from file
        # path_echarts = os.path.join(os.path.abspath("."), "core", "network_graph_data", "echarts_min.js")
        path_echarts = files("aiphoria.core").joinpath("network_graph_data/echarts_min.js")
        with open(path_echarts, mode="r", encoding="utf-8") as fs:
            self._echarts = fs.read()

        # Load visualizer script from file
        # path_network_graph = os.path.join(os.path.abspath("."), "core", "network_graph_data", "network_graph.js")
        path_network_graph = files("aiphoria.core").joinpath("network_graph_data/network_graph.js")
        with open(path_network_graph, mode="r", encoding="utf-8") as fs:
            self._visualizer = fs.read()

    def build(self, scenario_data: ScenarioData = None, options: Dict[str, Any] = None) -> None:
        """
        Compile and insert data to HTML file.

        :param scenario_data: ScenarioData object
        :param options: Extra options to be provided to network graph
        """
        if scenario_data is None:
            scenario_data = {}

        if options is None:
            options = {}

        year_to_process_id_to_process = scenario_data.year_to_process_id_to_process
        year_to_process_id_to_flow_ids = scenario_data.year_to_process_id_to_flow_ids
        year_to_flow_id_to_flow = scenario_data.year_to_flow_id_to_flow

        year_to_data = {}
        for year, process_id_to_process in year_to_process_id_to_process.items():
            node_index_to_data = {}
            edge_index_to_data = {}

            # Create nodes and edges
            new_node_index = 0
            new_edge_index = 0
            for process_id, process in process_id_to_process.items():
                inflow_ids = year_to_process_id_to_flow_ids[year][process_id]["in"]
                outflow_ids = year_to_process_id_to_flow_ids[year][process_id]["out"]

                # Create node data
                new_node_data = {
                    "process_id": process.id,
                    "process_label": process.label_in_graph,
                    "num_inflows": len(inflow_ids),
                    "num_outflows": len(outflow_ids),
                    "transformation_stage": process.transformation_stage,

                    # Stock related
                    "is_stock": process.stock_lifetime > 0,
                    "stock_lifetime": process.stock_lifetime,
                    "stock_distribution_type": process.stock_distribution_type,
                    "stock_distribution_params": process.stock_distribution_params,

                    # Virtual flow
                    "is_virtual": process.is_virtual
                }
                node_index_to_data[new_node_index] = new_node_data
                new_node_index += 1

                # Create edges from nodes (= process outflows)
                for flow_id in outflow_ids:
                    flow = year_to_flow_id_to_flow[year][flow_id]
                    new_edge_data = {
                        "flow_id": flow.id,
                        "source_process_id": flow.source_process_id,
                        "target_process_id": flow.target_process_id,
                        "is_unit_absolute_value": flow.is_unit_absolute_value,
                        "value": flow.value,  # Flow share
                        "unit": flow.unit,  # Flow unit
                        "evaluated_value": flow.evaluated_value,
                        "evaluated_share": flow.evaluated_share,

                        "is_virtual": flow.is_virtual,
                        "indicators": {k: v for k, v in flow.indicator_name_to_evaluated_value.items()},
                    }
                    edge_index_to_data[new_edge_index] = new_edge_data
                    new_edge_index += 1

            year_to_data[year] = {
                "node_index_to_data": node_index_to_data,
                "edge_index_to_data": edge_index_to_data,
            }

        # Build graph scenario data: contains data that is used shared between all processes and flows
        transformation_stage_name_to_color = {}
        if "transformation_stage_name_to_color" in options:
            transformation_stage_name_to_color = options["transformation_stage_name_to_color"]

        graph_scenario_data = {
            "scenario_name": options["scenario_name"],
            "baseline_unit_name": scenario_data.baseline_unit_name,
            "baseline_value_name": scenario_data.baseline_value_name,
            "transformation_stage_name_to_color": transformation_stage_name_to_color,
        }

        # Replace data in visualizer
        script = self._visualizer
        script = script.replace("{year_to_data}", json.dumps(year_to_data))
        script = script.replace("{scenario_data}", json.dumps(graph_scenario_data))
        self._visualizer = script

        # path_network_graph = os.path.join(os.path.abspath("."), "core", "network_graph_data", "network_graph.html")
        path_network_graph = files("aiphoria.core").joinpath("network_graph_data/network_graph.html")

        with open(path_network_graph, "r", encoding="utf-8") as fs:
            self._html = fs.read()

        # Replace 'echarts_content' and 'visualizer
        self._html = self._html.replace("{echarts_content}", self._echarts)
        self._html = self._html.replace("{visualizer_content}", self._visualizer)

    def show(self, output_filename: str = "network_graph_data.html") -> None:
        """
        Build HTML file and open it in browser.

        :param output_filename: Filename for the HTML file
        """
        with open(output_filename, "w", encoding="utf-8") as fs:
            fs.write(self._html)
        webbrowser.open("file://" + os.path.realpath(output_filename))



from enum import Enum


class VisualizerParameters(str, Enum):
    SmallNodeThreshold = "small_node_threshold"
    FlowAlpha = "flow_alpha"
    TransformationStageNameToColor = "transformation_stage_name_to_color"



import datetime
import sys
import time

# Logs
global_log_level = "info"

# Performance logs
use_log_perf = False
global_perf_logs = []
global_perf_log_index = 0


class bcolors(object):
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def log(msg: str, level: str = "info", show_timestamp: bool = False):
    """
    Log message with custom text.

    :param msg: Custom text to show
    :param level: Log level "info" (stdout) / "debug" (stderr)
    :param show_timestamp: True to show timestamp in message
    :return:
    """
    color = bcolors.OKGREEN

    prefix = ""
    target = sys.stdout
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    if level == "info":
        prefix = "INFO {}: " if show_timestamp else "INFO: "
        target = sys.stdout
        color = bcolors.OKGREEN

    if level == 'error':
        prefix = "ERROR: {}" if show_timestamp else "ERROR: "
        target = sys.stdout
        color = bcolors.FAIL

    if level == "debug":
        prefix = "DEBUG {}: " if show_timestamp else "INFO: "
        target = sys.stderr
        color = bcolors.FAIL

    if show_timestamp:
        prefix = prefix.format(timestamp)

    if not len(msg):
        prefix = ""

    text = prefix + msg
    print(color + text + bcolors.ENDC, file=target)
    target.flush()


def start_log_perf(name: str):
    if not use_log_perf:
        return

    global global_perf_logs
    global global_perf_log_index
    entry = [name, time.perf_counter()]
    global_perf_logs.append(entry)


def stop_log_perf():
    if not use_log_perf:
        return

    global global_perf_logs
    global global_perf_log_index
    time_in_secs = time.perf_counter() - global_perf_logs[global_perf_log_index][1]
    global_perf_logs[global_perf_log_index][1] = time_in_secs
    global_perf_log_index += 1
    return time_in_secs


def get_all_log_perf():
    for entry in global_perf_logs:
        yield entry


def show_log_perf_summary(title: str):
    if not use_log_perf:
        return

    target = sys.stderr
    target.flush()
    print("SUMMARY: {}".format(title), file=target)
    total_in_secs = sum([entry[1] for entry in global_perf_logs])
    for entry in global_perf_logs:
        name = entry[0]
        time_in_secs = entry[1]
        share = (time_in_secs / total_in_secs) * 100
        print("\t{}: {:.2f}s ({:.1f}%)".format(name, time_in_secs, share), file=target)
    target.flush()


def clear_log_perf():
    global global_perf_logs
    global global_perf_log_index
    global_perf_logs = []
    global_perf_log_index = 0



"""
    aiphoria core module
"""

from .builder import (
    init_builder,
    build_results,
    build_dataprovider,
    build_datachecker,
    build_and_solve_scenarios,
)

from .dataprovider import DataProvider
from .datachecker import DataChecker
from .datastructures import (
    Scenario,
    ScenarioData,
    Process,
    Flow,
    Stock,
    Indicator,
)

from .flowsolver import FlowSolver
from .parameters import ParameterName, ParameterFillMethod, StockDistributionType
from .datavisualizer import DataVisualizer
from .network_graph import NetworkGraph
from .logger import log
from .utils import (
    create_output_directory,
)

__all__ = [
    "init_builder",
    "build_results",
    "build_dataprovider",
    "build_datachecker",
    "build_and_solve_scenarios",
    "DataProvider",
    "DataChecker",
    "Scenario",
    "ScenarioData",
    "Process",
    "Flow",
    "Stock",
    "Indicator",
    "FlowSolver",
    "DataVisualizer",
    "NetworkGraph",
    "ParameterName",
    "ParameterFillMethod",
    "StockDistributionType",
    "log",
    "create_output_directory",
]



import os
import sys
import time
import shutil
import pandas as pd
import numpy as np
from tqdm import tqdm
from typing import Union
import matplotlib.pyplot as plt
from .core.builder import init_builder, build_results
from .core.utils import (
    setup_scenario_output_directories,
    calculate_scenario_mass_balance,
    shorten_sheet_name,
)
from .core.logger import log
from .core.parameters import ParameterName
from .core.network_graph import NetworkGraph
from .core.datavisualizer import DataVisualizer

_default_output_dir_name = "output"
_default_cache_dir_name = "cache"


def run_scenarios(path_to_settings_file: Union[str, None] = None,
                  path_to_output_dir: Union[str, None] = None,
                  remove_existing_output_dir: bool = False,
                  ) -> bool:
    """
    Run scenarios using the settings file.
    If path_to_output_dir is set then overrides the setting from Excel file.

    :param path_to_settings_file: Path to target settings Excel file
    :param path_to_output_dir: Path to output directory
    :param remove_existing_output_dir: Remove existing directory (default: False)
    """

    if path_to_settings_file is None:
        sys.stderr.write("ERROR: No path to settings file\n")
        sys.stderr.flush()
        return False

    abs_output_dir = os.path.realpath(os.path.expanduser(path_to_output_dir))
    output_dir_exists = os.path.isdir(abs_output_dir)
    if not remove_existing_output_dir and output_dir_exists:
        sys.stderr.write("ERROR: Output directory already exists\n")
        sys.stderr.flush()
        return False

    # Use the output path from settings file
    if path_to_output_dir is None:
        cwd = os.path.realpath(os.getcwd())
        path_to_output_dir = os.path.join(cwd, _default_output_dir_name)

    if os.path.isdir(path_to_output_dir):
        if not remove_existing_output_dir:
            log("Directory {} already exists.".format(path_to_output_dir))
            return False
        else:
            shutil.rmtree(path_to_output_dir, ignore_errors=True)

    if not os.path.isdir(path_to_output_dir):
        os.mkdir(path_to_output_dir)

    time_total_in_secs: float = time.perf_counter()
    path_to_cache = os.path.join(path_to_output_dir, _default_cache_dir_name)
    init_builder(path_to_cache=path_to_cache,
                 use_cache=False,
                 use_timing=False,
                 clear_cache=False)

    # Build results
    model_params, scenarios, color_definitions = build_results(path_to_settings_file, path_to_output_dir)

    scenario_name_to_output_path = setup_scenario_output_directories(
        model_params[ParameterName.OutputPath],
        [scenario.name for scenario in scenarios]
    )

    if model_params[ParameterName.CreateNetworkGraphs]:
        progress_bar = tqdm(total=len(scenarios),
                            desc="Building network graphs for solved scenarios")
        for scenario_index, scenario in enumerate(scenarios):
            # Extra options that are used when building network graphs
            options = {
                "transformation_stage_name_to_color": color_definitions,
                "scenario_name": scenario.name
            }

            output_filename = os.path.join(
                scenario_name_to_output_path.get(scenario.name), "network_graph.html")
            network_visualizer = NetworkGraph()
            network_visualizer.build(scenario.scenario_data, options)

            if model_params[ParameterName.ShowPlots]:
                network_visualizer.show(output_filename)

            progress_bar.update()
        progress_bar.close()
        sys.stdout.flush()
        sys.stderr.flush()

    # %%
    # ***************************************************************************
    # * Step 3: Export scenario results to files (processes, flows, and stocks) *
    # ***************************************************************************

    # Sheet names to what are written to file. Note that the order is important.
    sheet_names = ["Processes", "Flows",
                   "Flow values (baseline value)", "Mass balance"]
    sheet_name_to_list_of_dfs = {name: [] for name in sheet_names}

    progress_bar = tqdm(total=len(scenarios))
    for scenario_index, scenario in enumerate(scenarios):
        progress_bar.set_description("Exporting scenario processes, flows, and stocks data (scenario {}/{})".format(
            scenario_index + 1, len(scenarios)))

        # Processes Sheet
        df_processes = scenario.flow_solver.get_processes_as_dataframe()
        df_processes.insert(0, "Scenario", scenario.name)
        sheet_name_to_list_of_dfs[sheet_names[0]].append(df_processes)

        # Flows Sheet
        df_flows = scenario.flow_solver.get_flows_as_dataframe()
        df_flows.insert(0, "Scenario", scenario.name)
        sheet_name_to_list_of_dfs[sheet_names[1]].append(df_flows)

        # Flow values Sheet
        df_flow_values = scenario.flow_solver.get_evaluated_flow_values_as_dataframe()
        df_flow_values.insert(0, "Scenario", scenario.name)
        sheet_name_to_list_of_dfs[sheet_names[2]].append(df_flow_values)

        # Mass balance Sheet
        df_scenario_mass_balance = calculate_scenario_mass_balance(
            scenario.mfa_system)
        df_scenario_mass_balance.insert(0, "Scenario", scenario.name)
        sheet_name_to_list_of_dfs[sheet_names[3]].append(
            df_scenario_mass_balance)

        progress_bar.update(1)
    progress_bar.close()
    sys.stderr.flush()
    sys.stdout.flush()

    # Combine all scenario data to one Excel file
    # by concatenating all sheet-specific list of DataFrames as one DataFrame
    combined_excel_filename = os.path.join(
        model_params[ParameterName.OutputPath], "combined_scenario_data.xlsx")
    log(f"Exporting all scenarios to {combined_excel_filename}...")
    with pd.ExcelWriter(combined_excel_filename, engine='xlsxwriter') as writer:
        for sheet_name, list_of_dfs in sheet_name_to_list_of_dfs.items():
            df = pd.concat(list_of_dfs, ignore_index=True)
            df.to_excel(writer, sheet_name=sheet_name, index=False)

    log(f"All scenario data exported to {combined_excel_filename}")
    sys.stdout.flush()

    # %%
    # ***********************************************************************
    # * Step 4: Build dynamic stock results for each Scenario and visualize *
    # ***********************************************************************
    progress_bar = tqdm(total=len(scenarios),
                        desc="Building dynamic stock results")
    sys.stderr.flush()
    for scenario_index, scenario in enumerate(scenarios):
        progress_bar.set_description("Building dynamic stock results (scenario {}/{})".format(
            scenario_index + 1, len(scenarios)))

        flow_solver = scenario.flow_solver
        years = scenario.scenario_data.years
        scenario_output_path = scenario_name_to_output_path[scenario.name]

        # Full name of the baseline, e.g. "Solid wood equivalent"
        baseline_value_name = scenario.scenario_data.baseline_value_name
        baseline_unit_name = scenario.scenario_data.baseline_unit_name

        # Total number of indicators
        indicators = flow_solver.get_indicator_name_to_indicator()
        num_indicators = len(indicators.keys())

        # Baseline DSM
        stock_id_to_baseline_dsm = flow_solver.get_baseline_dynamic_stocks()
        stock_id_to_indicator_name_to_dsm = flow_solver.get_indicator_dynamic_stocks()

        if not len(stock_id_to_baseline_dsm.keys()):
            log("Scenario '{}': no dynamic stocks in the defined system".format(
                scenario.name))
            progress_bar.update(1)
            continue

        # Each baseline/indicator needs 3 plots
        # so total = baseline (3) + (number of indicators * 3)
        num_subplots = 3 + (num_indicators * 3)
        fig, axes = plt.subplots(
            num_subplots, 1, sharex='all', sharey='none', figsize=(12, 20))

        # Create an Excel writer for exporting data
        excel_filename = os.path.join(scenario_output_path, "{}_dynamic_stocks.xlsx".format(scenario.name))
        with pd.ExcelWriter(excel_filename, engine="xlsxwriter") as writer:
            all_stock_total_dfs = []
            all_stock_change_dfs = []
            all_stock_outflow_dfs = []
            for stock_id, baseline_dsm in stock_id_to_baseline_dsm.items():
                plot_index = 0

                # Truncate the stock ID to 20 characters (or any suitable length) to fit within the 31 character limit
                # Truncate to the first 20 characters
                stock_id_for_filename = stock_id[:20]
                stock_id_for_filename = stock_id_for_filename.replace(
                    ":", "_")  # Replace ":" with "_"

                # ******************
                # * Baseline stock *
                # ******************
                baseline_stock_by_cohort = baseline_dsm.compute_s_c_inflow_driven()
                baseline_outflow_by_cohort = baseline_dsm.compute_o_c_from_s_c()
                baseline_stock_total = baseline_dsm.compute_stock_total()
                baseline_stock_change = baseline_dsm.compute_stock_change()
                baseline_stock_outflow = baseline_dsm.compute_outflow_total()

                # Export stock by cohort
                sheet_name = shorten_sheet_name(f'{stock_id_for_filename}_s_by_c_{baseline_value_name}')
                df_baseline_stock_by_cohort = pd.DataFrame(
                    baseline_stock_by_cohort, columns=years, index=years)
                df_baseline_stock_by_cohort.to_excel(
                    writer, sheet_name=sheet_name)

                # Export outflow by cohort
                sheet_name = shorten_sheet_name(f'{stock_id_for_filename}_o_by_c_{baseline_value_name}')
                df_baseline_outflow_by_cohort = pd.DataFrame(
                    baseline_outflow_by_cohort, columns=years, index=years)
                df_baseline_outflow_by_cohort.to_excel(
                    writer, sheet_name=sheet_name)

                # Export stock total
                df_baseline_stock_total = pd.DataFrame(baseline_stock_total, index=years)
                df_baseline_stock_total.reset_index(inplace=True)
                df_baseline_stock_total.columns = ["Year", "Stock total"]
                df_baseline_stock_total["Scenario"] = scenario.name
                df_baseline_stock_total["Stock ID"] = stock_id
                df_baseline_stock_total["Indicator"] = baseline_unit_name
                all_stock_total_dfs.append(df_baseline_stock_total)

                # Export stock change
                df_baseline_stock_change = pd.DataFrame(baseline_stock_change, index=years)
                df_baseline_stock_change.reset_index(inplace=True)
                df_baseline_stock_change.columns = ["Year", "Stock change"]
                df_baseline_stock_change["Scenario"] = scenario.name
                df_baseline_stock_change["Stock ID"] = stock_id
                df_baseline_stock_change["Indicator"] = baseline_unit_name
                all_stock_change_dfs.append(df_baseline_stock_change)

                # Export stock outflow total
                df_baseline_stock_outflow = pd.DataFrame(baseline_stock_outflow, index=years)
                df_baseline_stock_outflow.reset_index(inplace=True)
                df_baseline_stock_outflow.columns = ["Year", "Stock outflow total"]
                df_baseline_stock_outflow["Scenario"] = scenario.name
                df_baseline_stock_outflow["Stock ID"] = stock_id
                df_baseline_stock_outflow["Indicator"] = baseline_unit_name
                all_stock_outflow_dfs.append(df_baseline_stock_outflow)

                # Plot baseline stock total (in-use stocks)
                axes[plot_index + 0].plot(years, baseline_stock_total, marker='o', label="{}".format(stock_id))
                axes[plot_index + 0].set_ylabel("In-use stock ({})".format(baseline_unit_name))
                axes[plot_index + 0].set_title("In-use stock per year by product type")

                # Plot baseline stock change
                axes[plot_index + 1].plot(years, baseline_stock_change, marker='o', label=f'{stock_id}')
                axes[plot_index + 1].set_ylabel("Stock change ({})".format(baseline_unit_name))
                axes[plot_index + 1].set_title("Stock change per year by product type")

                # Plot baseline outflow by cohort
                axes[plot_index + 2].plot(years, baseline_stock_outflow, marker='o', label=f'{stock_id}')
                axes[plot_index + 2].set_ylabel("Stock outflow ({})".format(baseline_unit_name))
                axes[plot_index + 2].set_title("Stock outflow per year by product type")

                plot_index += 3
                for indicator_name, indicator_dsm in stock_id_to_indicator_name_to_dsm[stock_id].items():
                    # **************
                    # * Indicators *
                    # **************
                    indicator_unit = indicators[indicator_name].unit
                    indicator_stock_by_cohort = indicator_dsm.compute_s_c_inflow_driven()
                    indicator_outflow_by_cohort = indicator_dsm.compute_o_c_from_s_c()
                    indicator_stock_total = indicator_dsm.compute_stock_total()
                    indicator_stock_change = indicator_dsm.compute_stock_change()
                    indicator_stock_outflow = indicator_dsm.compute_outflow_total()

                    # Export indicator stock by cohort
                    sheet_name = shorten_sheet_name(f"{stock_id_for_filename}_s_by_c_{indicator_name}")
                    df_indicator_stock_by_cohort = pd.DataFrame(
                        indicator_stock_by_cohort, columns=years, index=years)
                    df_indicator_stock_by_cohort.to_excel(writer, sheet_name=sheet_name)

                    # Export indicator outflow by cohort
                    sheet_name = shorten_sheet_name(f"{stock_id_for_filename}_o_by_c_{indicator_name}")
                    df_indicator_oc = pd.DataFrame(
                        indicator_outflow_by_cohort, columns=years, index=years)
                    df_indicator_oc.to_excel(writer, sheet_name=sheet_name)

                    # Export indicator stock total
                    df_indicator_stock_total = pd.DataFrame(
                        indicator_stock_total, index=years)
                    df_indicator_stock_total.reset_index(inplace=True)
                    df_indicator_stock_total.columns = ["Year", "Stock total"]
                    df_indicator_stock_total["Scenario"] = scenario.name
                    df_indicator_stock_total["Stock ID"] = stock_id
                    df_indicator_stock_total["Indicator"] = indicator_name
                    all_stock_total_dfs.append(df_indicator_stock_total)

                    # Export indicator stock change
                    df_indicator_stock_change = pd.DataFrame(
                        indicator_stock_change, index=years)
                    df_indicator_stock_change.reset_index(inplace=True)
                    df_indicator_stock_change.columns = [
                        "Year", "Stock change"]
                    df_indicator_stock_change["Scenario"] = scenario.name
                    df_indicator_stock_change["Stock ID"] = stock_id
                    df_indicator_stock_change["Indicator"] = indicator_name
                    all_stock_change_dfs.append(df_indicator_stock_change)

                    # Export indicator stock outflow total
                    df_indicator_stock_outflow = pd.DataFrame(
                        indicator_stock_outflow, index=years)
                    df_indicator_stock_outflow.reset_index(inplace=True)
                    df_indicator_stock_outflow.columns = [
                        "Year", "Stock outflow total"]
                    df_indicator_stock_outflow["Scenario"] = scenario.name
                    df_indicator_stock_outflow["Stock ID"] = stock_id
                    df_indicator_stock_outflow["Indicator"] = indicator_name
                    all_stock_outflow_dfs.append(df_indicator_stock_outflow)

                    # Plot indicator stock total (in-use stocks)
                    axes[plot_index + 0].plot(years, indicator_stock_total, marker='o', label='{} ({}) {}'.format(
                        indicator_name, indicator_unit, stock_id))
                    axes[plot_index +
                         0].set_ylabel("In-use stock ({})".format(indicator_unit))
                    axes[plot_index + 0].set_title(
                        "{} stock in-use per year by product type".format(indicator_name))

                    # Plot indicator stock change
                    axes[plot_index + 1].plot(years, indicator_stock_change, marker='o', label="{} ({}) {}".format(
                        indicator_name, indicator_unit, stock_id))
                    axes[plot_index +
                         1].set_ylabel("Stock change ({})".format(indicator_unit))
                    axes[plot_index +
                         1].set_title("{} stock change per year".format(indicator_name))

                    # Plot indicator outflow by cohort
                    axes[plot_index + 2].plot(years, indicator_stock_outflow, marker='o', label="{} ({}) {}".format(
                        indicator_name, indicator_unit, stock_id
                    ))
                    axes[plot_index +
                         2].set_ylabel("Stock outflow ({})".format(indicator_unit))
                    axes[plot_index + 2].set_title(
                        "{} outflow per year by product type".format(indicator_name))

                    plot_index += 3

            if all_stock_total_dfs:
                combined_stock_total_df = pd.concat(
                    all_stock_total_dfs, ignore_index=True)
                combined_sheet_name = "Total_stock"
                combined_stock_total_df.to_excel(
                    writer, sheet_name=combined_sheet_name, index=False)

            if all_stock_change_dfs:
                combined_stock_change_df = pd.concat(
                    all_stock_change_dfs, ignore_index=True)
                combined_sheet_name = "Total_stock_change"
                combined_stock_change_df.to_excel(
                    writer, sheet_name=combined_sheet_name, index=False)

            if all_stock_outflow_dfs:
                all_stock_outflow_dfs = pd.concat(
                    all_stock_outflow_dfs, ignore_index=True)
                combined_sheet_name = "Total_stock_outflow"
                all_stock_outflow_dfs.to_excel(
                    writer, sheet_name=combined_sheet_name, index=False)

            # Set common properties to axes
            range_x_ticks = range(min(years), max(years) + 1)
            for axis in axes:
                axis.set_xlabel("Year")
                axis.title.set_size(12)
                axis.legend()

            # Adjust layout to prevent overlap
            plt.tight_layout()
            tick_gap = 1 if len(years) < 15 else 10
            plt.xticks(years[::tick_gap])

            # # Save the figure as an SVG file
            # filename = os.path.join(scenario_output_path, "{}_stock_plots_by_product.svg".format(scenario.name))
            # plt.savefig(filename, format='svg')
            #
            # if model_params[ParameterName.ShowPlots]:
            #     plt.show()

        progress_bar.update(1)
    progress_bar.refresh()
    progress_bar.close()
    sys.stderr.flush()

    # *****************************************************
    # * Step 5: Convert the carbon stocks to CO2 removals *
    # *****************************************************
    log("Calculating annual CO2 stock emissions / removals results...")

    # Storage for comparison
    all_scenario_results = {}
    all_emitter_years = {}

    show_steady_state_overlay = False  # Toggle to enable/disable overlay
    steady_state_threshold_ratio = 0.05  # Relative threshold for stability
    min_steady_state_years = 5  # Minimum consecutive years for valid steady state

    for scenario in scenarios:
        scenario_output_path = scenario_name_to_output_path[scenario.name]
        flow_solver = scenario.flow_solver
        years = scenario.scenario_data.years
        year_start = scenario.scenario_data.start_year

        stock_id_to_indicator_name_to_dsm = flow_solver.get_indicator_dynamic_stocks()
        if not len(stock_id_to_indicator_name_to_dsm.keys()):
            log("Scenario '{}': no dynamic stocks in the defined system".format(
                scenario.name))
            continue

        results_co2_removals = pd.DataFrame({'Year': years})
        results_net_emitters = pd.DataFrame({'Year': years})
        conversion_factor_c_to_co2 = model_params[ParameterName.ConversionFactorCToCO2]

        # Define line styles, markers, and colors for differentiation
        line_styles = ['-', '--', '-.', ':']
        markers = ['o', 's', '^', 'D']
        colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k']

        target_indicator_name = "Carbon"
        plt.figure(figsize=(10, 6))

        scenario_results = {}
        steady_state_info = {}
        net_emitter_info = {}

        for index, (stock_id, indicator_name_to_dsm) in enumerate(stock_id_to_indicator_name_to_dsm.items()):
            if target_indicator_name not in indicator_name_to_dsm:
                continue

            dsm = indicator_name_to_dsm[target_indicator_name]
            total_inflows_carbon = dsm.i
            total_outflows_carbon = dsm.o
            annual_co2_removal = (
                total_inflows_carbon - total_outflows_carbon) * conversion_factor_c_to_co2
            results_co2_removals[stock_id] = annual_co2_removal
            scenario_results[stock_id] = annual_co2_removal

            # Detect steady-state years with rolling window approach
            threshold = steady_state_threshold_ratio * \
                max(abs(annual_co2_removal))
            rolling_mean = pd.Series(annual_co2_removal).rolling(
                window=min_steady_state_years, center=True).mean()
            is_steady = abs(pd.Series(annual_co2_removal) -
                            rolling_mean) < threshold

            # Extract consecutive steady years
            steady_years = []
            current_run = []
            for year, steady in zip(years, is_steady):
                if steady:
                    current_run.append(year)
                else:
                    if len(current_run) >= min_steady_state_years:
                        steady_years.extend(current_run)
                    current_run = []
            if len(current_run) >= min_steady_state_years:
                steady_years.extend(current_run)

            steady_state_info[stock_id] = sorted(set(steady_years))

            # Detect net emitter years (negative removals)
            emitter_years = [year for year, value in zip(
                years, annual_co2_removal) if value < 0]
            results_net_emitters[stock_id] = [
                "Emitter" if value < 0 else "" for value in annual_co2_removal]
            net_emitter_info[stock_id] = emitter_years

            # Plot CO2 removals with steady state overlay
            line_style = line_styles[index % len(line_styles)]
            marker = markers[index % len(markers)]
            color = colors[index % len(colors)]
            plt.plot(years, annual_co2_removal, marker=marker, linestyle=line_style, color=color,
                     label=f'{stock_id}')
            if show_steady_state_overlay and steady_years:
                plt.axvspan(steady_years[0],
                            steady_years[-1], color=color, alpha=0.1)

        all_scenario_results[scenario.name] = scenario_results
        all_emitter_years[scenario.name] = net_emitter_info

        plt.xlabel('Year')
        plt.ylabel('CO2 Emissions / Removals (Mt CO2)')
        plt.title('Annual CO2 Emissions / Removals by Product')
        plt.grid(True)
        tick_gap = 1 if len(years) < 15 else 10
        plt.xticks(years[::tick_gap])
        plt.legend()
        plt.tight_layout()

        # Export CO2 removal data to CSV
        log("Exporting annual CO2 emissions / removal (Mt) by stock results...")
        filename = os.path.join(scenario_output_path, f"{scenario.name}_annual_co2_removal_by_stock.csv")
        results_co2_removals.to_csv(path_or_buf=filename, index=False, mode="w")

        # Export net emitter flag table
        log("Exporting annual CO2 net emitter years (where removals < 0)...")
        filename = os.path.join(scenario_output_path, f"{scenario.name}_annual_net_emitter_flags.csv")
        results_net_emitters.to_csv(path_or_buf=filename, index=False, mode="w")

        # Export CO2 removal plot as SVG
        filename = os.path.join(scenario_output_path, f"{scenario.name}_annual_co2_removal_by_product.svg")
        plt.savefig(filename, format='svg')

        # if model_params[ParameterName.ShowPlots]:
        #     plt.show()

        # Print and export steady-state info
        print(f"\nSteady-state periods for scenario '{scenario.name}':")
        for stock_id, years_list in steady_state_info.items():
            if years_list:
                print(f"  {stock_id}: {years_list[0]} to {years_list[-1]} ({len(years_list)} years)")
            else:
                print(f"  {stock_id}: No steady-state period detected.")

        steady_state_df = pd.DataFrame([
            {'Stock': stock_id, 'StartYear': years_list[0] if years_list else None,
             'EndYear': years_list[-1] if years_list else None, 'DurationYears': len(years_list)}
            for stock_id, years_list in steady_state_info.items()
        ])
        filename = os.path.join(scenario_output_path, f"{scenario.name}_steady_state_periods.csv")
        steady_state_df.to_csv(filename, index=False)

    # ************************************************************
    # * Step 6: Visualize inflows per year to selected processes *
    # ************************************************************

    # This is only done when there is multiple years
    # Visualize inflows per year to processes
    visualize_inflows_to_process_ids = model_params[ParameterName.VisualizeInflowsToProcesses]
    for scenario in scenarios:
        scenario_output_path = scenario_name_to_output_path[scenario.name]
        flow_solver = scenario.flow_solver
        years = scenario.scenario_data.years

        # Dictionary: Process ID to process
        unique_processes = flow_solver.get_unique_processes()
        for process_id in visualize_inflows_to_process_ids:
            process = flow_solver.get_process(process_id, min(years))
            flow_id_to_source_process_id = {}

            # Find all source processes of all incoming flows to this process in all years
            # This is needed to create stable set of process names so that the relative
            # position of the processes stay the same in stacked chart between the years
            source_process_ids = set()
            for year in years:
                inflows = flow_solver.get_process_flows(
                    process_id, year)["Inflows"]
                unique_flow_ids = set()
                for flow in inflows:
                    unique_flow_ids.add(flow.id)
                    flow_id_to_source_process_id[flow.id] = flow.source_process_id

                # Find source process ID of each incoming flow and add
                # to list of unique source process IDs if not already there
                unique_flow_ids = list(unique_flow_ids)
                for flow_id in unique_flow_ids:
                    source_process_ids.add(
                        flow_id_to_source_process_id[flow_id])

            # Now source_process_ids-list contains list of all the possible process IDs
            # that have flows incoming to process_id. This list is needed to keep the
            # incoming process IDs the same every year because aiphoria allows the connections
            # between the flows to change between the years.
            source_process_ids = list(source_process_ids)

            # Create 2D array with shape of (number of source process IDs, number of years)
            # and fill with the value of the inflow from source process for each year
            df_inflows_to_process = pd.DataFrame(columns=['Year', 'Source Process ID', 'Value ({})'.format(
                model_params[ParameterName.BaselineUnitName])])
            source_process_by_flow_values = np.zeros(
                (len(source_process_ids), len(years)))
            for year_index, year in enumerate(years):
                inflows = flow_solver.get_process_flows(
                    process_id, year)["Inflows"]
                for flow in inflows:
                    source_process_id_index = source_process_ids.index(
                        flow.source_process_id)
                    source_process_by_flow_values[source_process_id_index,
                                                  year_index] = flow.evaluated_value
                    df_inflows_to_process.loc[len(df_inflows_to_process)] = [year, flow.source_process_id,
                                                                             flow.evaluated_value]

            df_inflows_to_process = df_inflows_to_process.round(5)

            # Export inflows to process to CSV file
            # NOTE: Replace character ':' in Process ID to underscore because
            # Windows system are not able to handle that character in filename
            process_id_for_filename = process_id.replace(":", "_")
            filename = os.path.join(scenario_output_path,
                                    "{}_inflows_to_{}.csv".format(scenario.name, process_id_for_filename))
            df_inflows_to_process.to_csv(
                path_or_buf=filename, index=False, mode="w")

            # Initialize the figure and axes for the stacked area chart
            fig, ax = plt.subplots(figsize=(12, 8))
            ax.stackplot(years, source_process_by_flow_values,
                         labels=list(source_process_ids))
            ax.set_ylabel("Mm3 SWE")
            ax.set_title("Inputs to {}".format(process.name))
            ax.legend(loc='upper left')
            tick_gap = 1 if len(years) < 15 else 10
            plt.xticks(years[::tick_gap])

            # Save the figure as an SVG file
            filename = os.path.join(scenario_output_path,
                                    "{}_inflows_to_{}.svg".format(scenario.name, process_id_for_filename))
            plt.savefig(filename, format='svg')

            # NOTE: Causes when running from PyCharm
            # if model_params[ParameterName.ShowPlots]:
            #     plt.show()

    # ***********************************************************
    # * Step 7: Visualize the scenario results as Sankey graphs *
    # ***********************************************************
    # Virtual process graph label overrides
    # TODO: Move also this to settings file?
    virtual_process_graph_labels = {}
    virtual_process_graph_labels["VP_P2:EU"] = "Unreported flow from P2"
    virtual_process_graph_labels["VP_P3:EU"] = "Unreported flow from P3"

    # Virtual Process and virtual Flow colors
    visualizer_params = {
        # User can hide processes in Sankey graph that have total inflows less than this value
        # This value cannot be changed now in the Sankey graph
        # TODO: Move this to settings file?
        "small_node_threshold": 5,

        # Dictionary to define labels for virtual flows
        # If dictionary contains label for the virtual process then that is used,
        # otherwise the virtual process ID is used
        "virtual_process_graph_labels": virtual_process_graph_labels,

        # Dictionary to define color of process by the process transformation stage name
        # All must be provided as a RGB hex string, prefixed by character '#'
        # Usage example: { "Source": "#707070" }
        "process_transformation_stage_colors": color_definitions,

        # How transparent flows are (0.0 = invisible, 1.0 = fully opaque)
        "flow_alpha": 0.75,

        # Color for virtual process
        "virtual_process_color": "rgba(0.3, 0.3, 0.3, 0.6)",

        # Color for virtual flows
        "virtual_flow_color": "#808080",

        # NOTE: Data inside metadata is automatically embedded to Sankey visualization
        "metadata": {},
    }

    if model_params[ParameterName.IncludeMetadata]:
        visualizer_params["metadata"] = {
            "filename": os.path.basename(path_to_settings_file)
        }

    if model_params[ParameterName.CreateSankeyCharts]:
        log("Creating Sankey charts for scenarios...")
        visualizer = DataVisualizer()
        visualizer.build_and_show(
            scenarios, visualizer_params, model_params, combine_to_one_file=True)

    time_total_in_secs = time.perf_counter() - time_total_in_secs
    log("Finished in {:.2f}s".format(time_total_in_secs))
